"""Robust and Unbiased Model-BAsed Spherical Deconvolution (RUMBA-SD)"""

import logging
import warnings

import numpy as np

from dipy.core.geometry import vec2vec_rotmat
from dipy.core.gradients import get_bval_indices, gradient_table, unique_bvals_tolerance
from dipy.core.onetime import auto_attr
from dipy.core.sphere import Sphere
from dipy.data import get_sphere
from dipy.reconst.csdeconv import AxSymShResponse
from dipy.reconst.odf import OdfFit, OdfModel
from dipy.reconst.shm import lazy_index, normalize_data
from dipy.segment.mask import bounding_box, crop
from dipy.sims.voxel import all_tensor_evecs, single_tensor
from dipy.testing.decorators import warning_for_keywords

# Machine precision for numerical stability in division
_EPS = np.finfo(float).eps
logger = logging.getLogger(__name__)


class RumbaSDModel(OdfModel):
    @warning_for_keywords()
    def __init__(
        self,
        gtab,
        *,
        wm_response=(1.7e-3, 0.2e-3, 0.2e-3),
        gm_response=0.8e-3,
        csf_response=3.0e-3,
        n_iter=600,
        recon_type="smf",
        n_coils=1,
        R=1,
        voxelwise=True,
        use_tv=False,
        sphere=None,
        verbose=False,
    ):
        """
        Robust and Unbiased Model-BAsed Spherical Deconvolution (RUMBA-SD).

        RUMBA-SD :footcite:p:`CanalesRodriguez2015` is a modification of the
        Richardson-Lucy algorithm accounting for Rician and Noncentral Chi noise
        distributions, which more accurately represent MRI noise. Computes a
        maximum likelihood estimation of the fiber orientation density function
        (fODF) at each voxel. Includes white matter compartments alongside
        optional GM and CSF compartments to account for partial volume effects.
        This fit can be performed voxelwise or globally. The global fit will
        proceed more quickly than the voxelwise fit provided that the computer
        has adequate RAM (>= 16 GB should be sufficient for most datasets).

        Kernel for deconvolution constructed using a priori knowledge of white
        matter response function, as well as the mean diffusivity of GM and/or
        CSF. RUMBA-SD is robust against impulse response imprecision, and thus
        the default diffusivity values are often adequate
        :footcite:p:`DellAcqua2007`.


        Parameters
        ----------
        gtab : GradientTable
            Gradient table.
        wm_response : 1d ndarray or 2d ndarray or AxSymShResponse, optional
            Tensor eigenvalues as a (3,) ndarray, multishell eigenvalues as
            a (len(unique_bvals_tolerance(gtab.bvals))-1, 3) ndarray in
            order of smallest to largest b-value, or an AxSymShResponse.
            Default: np.array([1.7e-3, 0.2e-3, 0.2e-3])
        gm_response : float, optional
            Mean diffusivity for GM compartment. If `None`, then grey
            matter volume fraction is not computed. Default: 0.8e-3
        csf_response : float, optional
            Mean diffusivity for CSF compartment. If `None`, then CSF
            volume fraction is not computed. Default: 3.0e-3
        n_iter : int, optional
            Number of iterations for fODF estimation. Must be a positive int.
            Default: 600
        recon_type : {'smf', 'sos'}, optional
            MRI reconstruction method: spatial matched filter (SMF) or
            sum-of-squares (SoS). SMF reconstruction generates Rician noise
            while SoS reconstruction generates Noncentral Chi noise.
            Default: 'smf'
        n_coils : int, optional
            Number of coils in MRI scanner -- only relevant in SoS
            reconstruction. Must be a positive int. Default: 1
        R : int, optional
            Acceleration factor of the acquisition. For SIEMENS,
            R = iPAT factor. For GE, R = ASSET factor. For PHILIPS,
            R = SENSE factor. Typical values are 1 or 2. Must be a positive
            int. Default: 1
        voxelwise : bool, optional
            If true, performs a voxelwise fit. If false, performs a global fit
            on the entire brain at once. The global fit requires a 4D brain
            volume in `fit`.
        use_tv : bool, optional
            If true, applies total variation regularization. This only takes
            effect in a global fit (`voxelwise` is set to `False`). TV can only
            be applied to 4D brain volumes with no singleton dimensions.
        sphere : Sphere, optional
            Sphere on which to construct fODF. If None, uses `repulsion724`.
        verbose : bool, optional
            If true, logs updates on estimated signal-to-noise ratio after each
            iteration. This only takes effect in a global fit (`voxelwise` is
            set to `False`).

        References
        ----------
        .. footbibliography::

        """

        if not np.any(gtab.b0s_mask):
            raise ValueError("Gradient table has no b0 measurements")

        self.gtab_orig = gtab  # save for prediction

        # Masks to extract b0/non-b0 measurements
        self.where_b0s = lazy_index(gtab.b0s_mask)
        self.where_dwi = lazy_index(~gtab.b0s_mask)

        # Correct gradient table to contain b0 data at the beginning
        bvals_cor = np.concatenate(([0], gtab.bvals[self.where_dwi]))
        bvecs_cor = np.concatenate(([[0, 0, 0]], gtab.bvecs[self.where_dwi]))
        gtab_cor = gradient_table(bvals_cor, bvecs=bvecs_cor)

        # Initialize self.gtab
        OdfModel.__init__(self, gtab_cor)

        if isinstance(wm_response, tuple):
            wm_response = np.asarray(wm_response)

        # Store responses
        self.wm_response = wm_response
        self.gm_response = gm_response
        self.csf_response = csf_response

        # Initializing remaining parameters
        if R < 1 or n_iter < 1 or n_coils < 1:
            raise ValueError(
                f"R, n_iter, and n_coils must be >= 1, but R={R},"
                + f"n_iter={n_iter}, and n_coils={n_coils} "
            )

        self.R = R
        self.n_iter = n_iter
        self.recon_type = recon_type
        self.n_coils = n_coils

        if voxelwise and use_tv:
            raise ValueError("Total variation has no effect in voxelwise fit")
        if voxelwise and verbose:
            warnings.warn(
                "Verbosity has no effect in voxelwise fit", UserWarning, stacklevel=2
            )

        self.voxelwise = voxelwise
        self.use_tv = use_tv

        self.verbose = verbose

        if sphere is None:
            self.sphere = get_sphere(name="repulsion724")
        else:
            self.sphere = sphere

        if voxelwise:
            self.fit = self._voxelwise_fit
        else:
            self.fit = self._global_fit

        # Fitting parameters
        self.kernel = None

    @warning_for_keywords()
    def _global_fit(self, data, *, mask=None):
        """
        Fit fODF and GM/CSF volume fractions globally.

        Parameters
        ----------
        data : ndarray (x, y, z, N)
            Signal values for each voxel. Must be 4D.
        mask : ndarray (x, y, z), optional
            Binary mask specifying voxels of interest with 1; results will only
            be fit at these voxels (0 elsewhere). If `None`, fits all voxels.
            Default: None.

        Returns
        -------
        model_fit : RumbaFit
            Fit object storing model parameters.

        """

        # Checking data and mask shapes
        if len(data.shape) != 4:
            raise ValueError(f"Data should be 4D, received shape f{data.shape}")

        if mask is None:  # default mask includes all voxels
            mask = np.ones(data.shape[:3])

        if data.shape[:3] != mask.shape:
            raise ValueError(
                "Mask shape should match first 3 dimensions of "
                + f"data, but data dimensions are f{data.shape} "
                + f"while mask dimensions are f{mask.shape}"
            )

        # Signal repair, normalization

        # Normalize data to mean b0 image
        data = normalize_data(data, self.where_b0s, min_signal=_EPS)
        # Rearrange data to match corrected gradient table
        data = np.concatenate(
            (np.ones([*data.shape[:3], 1]), data[..., self.where_dwi]), axis=3
        )
        data[data > 1] = 1  # clip values between 0 and 1

        # All arrays are converted to float32 to reduce memory load
        data = data.astype(np.float32)

        # Generate kernel
        self.kernel = generate_kernel(
            self.gtab,
            self.sphere,
            self.wm_response,
            self.gm_response,
            self.csf_response,
        ).astype(np.float32)

        # Fit fODF
        model_params = rumba_deconv_global(
            data,
            self.kernel,
            mask,
            n_iter=self.n_iter,
            recon_type=self.recon_type,
            n_coils=self.n_coils,
            R=self.R,
            use_tv=self.use_tv,
            verbose=self.verbose,
        )

        model_fit = RumbaFit(self, model_params)
        return model_fit

    @warning_for_keywords()
    def _voxelwise_fit(self, data, *, mask=None):
        """
        Fit fODF and GM/CSF volume fractions voxelwise.

        Parameters
        ----------
        data : ndarray ([x, y, z], N)
            Signal values for each voxel.
        mask : ndarray ([x, y, z]), optional
            Binary mask specifying voxels of interest with 1; results will only
            be fit at these voxels (0 elsewhere). If `None`, fits all voxels.
            Default: None.

        Returns
        -------
        model_fit : RumbaFit
            Fit object storing model parameters.

        """

        if mask is None:  # default mask includes all voxels
            mask = np.ones(data.shape[:-1])

        if data.shape[:-1] != mask.shape:
            raise ValueError(
                "Mask shape should match first dimensions of "
                + f"data, but data dimensions are f{data.shape} "
                + f"while mask dimensions are f{mask.shape}"
            )

        self.kernel = generate_kernel(
            self.gtab,
            self.sphere,
            self.wm_response,
            self.gm_response,
            self.csf_response,
        )

        model_params = np.zeros(data.shape[:-1] + (len(self.sphere.vertices) + 2,))

        for ijk in np.ndindex(data.shape[:-1]):
            if mask[ijk]:
                vox_data = data[ijk]
                # Normalize data to mean b0 image
                vox_data = normalize_data(vox_data, self.where_b0s, min_signal=_EPS)
                # Rearrange data to match corrected gradient table
                vox_data = np.concatenate(([1], vox_data[self.where_dwi]))
                vox_data[vox_data > 1] = 1  # clip values between 0 and 1

                # Fitting
                model_param = rumba_deconv(
                    vox_data,
                    self.kernel,
                    n_iter=self.n_iter,
                    recon_type=self.recon_type,
                    n_coils=self.n_coils,
                )

                model_params[ijk] = model_param

        model_fit = RumbaFit(self, model_params)
        return model_fit


class RumbaFit(OdfFit):
    def __init__(self, model, model_params):
        """
        Constructs fODF, GM/CSF volume fractions, and other derived results.

        fODF and GM/CSF fractions are normalized to collectively sum to 1 for
        each voxel.

        Parameters
        ----------
        model : RumbaSDModel
            RumbaSDModel-SD model.
        model_params : ndarray ([x, y, z], M)
            fODF and GM/CSF volume fractions for each voxel.

        """

        self.model = model
        self.model_params = model_params

    @warning_for_keywords()
    def odf(self, *, sphere=None):
        """
        Constructs fODF at discrete vertices on model sphere for each voxel.

        Parameters
        ----------
        sphere : Sphere, optional
            Sphere on which to construct fODF. If specified, must be the same
            sphere used by the `RumbaSDModel` model. Default: None.

        Returns
        -------
        odf : ndarray ([x, y, z], M-2)
            fODF computed at each vertex on model sphere.

        """
        if sphere is not None and sphere != self.model.sphere:
            raise ValueError(
                "Reconstruction sphere must be the same as used"
                + " in the RUMBA-SD model."
            )

        odf = self.model_params[..., :-2]
        return odf

    @auto_attr
    def f_gm(self):
        """
        Constructs GM volume fraction for each voxel.

        Returns
        -------
        f_gm : ndarray ([x, y, z])
            GM volume fraction.
        """

        f_gm = self.model_params[..., -2]
        return f_gm

    @auto_attr
    def f_csf(self):
        """
        Constructs CSF volume fraction for each voxel.

        Returns
        -------
        f_csf : ndarray ([x, y, z])
            CSF volume fraction.
        """

        f_csf = self.model_params[..., -1]
        return f_csf

    @auto_attr
    def f_wm(self):
        """
        Constructs white matter volume fraction for each voxel.

        Equivalent to sum of fODF.

        Returns
        -------
        f_wm : ndarray ([x, y, z])
            White matter volume fraction.
        """

        f_wm = np.sum(self.odf(), axis=-1)
        return f_wm

    @auto_attr
    def f_iso(self):
        """
        Constructs isotropic volume fraction for each voxel.

        Equivalent to sum of GM and CSF volume fractions.

        Returns
        -------
        f_iso : ndarray ([x, y, z])
            Isotropic volume fraction.
        """

        f_iso = self.f_gm + self.f_csf
        return f_iso

    @auto_attr
    def combined_odf_iso(self):
        """
        Constructs fODF combined with isotropic volume fraction at discrete
        vertices on model sphere.

        Distributes isotropic compartments evenly along each fODF direction.
        Sums to 1.

        Returns
        -------
        combined : ndarray ([x, y, z], M-2)
            fODF combined with isotropic volume fraction.
        """

        odf = self.odf()
        combined = odf + self.f_iso[..., None] / odf.shape[-1]
        return combined

    @warning_for_keywords()
    def predict(self, *, gtab=None, S0=None):
        """
        Compute signal prediction on model gradient table given given fODF
        and GM/CSF volume fractions for each voxel.

        Parameters
        ----------
        gtab : GradientTable, optional
            The gradients for which the signal will be predicted. Use the
            model's gradient table if `None`. Default: None
        S0 : ndarray ([x, y, z]) or float, optional
            The non diffusion-weighted signal value for each voxel. If a float,
            the same value is used for each voxel. If `None`, 1 is used for
            each voxel. Default: None

        Returns
        -------
        pred_sig : ndarray ([x, y, z], N)
            The predicted signal.

        """
        model_params = self.model_params

        if gtab is None:
            gtab = self.model.gtab_orig

        pred_kernel = generate_kernel(
            gtab,
            self.model.sphere,
            self.model.wm_response,
            self.model.gm_response,
            self.model.csf_response,
        )

        if S0 is None:
            S0 = np.ones(model_params.shape[:-1] + (1,))
        elif isinstance(S0, np.ndarray):
            S0 = S0[..., None]
        else:
            S0 = S0 * np.ones(model_params.shape[:-1] + (1,))

        pred_sig = np.empty(model_params.shape[:-1] + (len(gtab.bvals),))
        for ijk in np.ndindex(model_params.shape[:-1]):
            pred_sig[ijk] = S0[ijk] * np.dot(pred_kernel, model_params[ijk])

        return pred_sig


@warning_for_keywords()
def rumba_deconv(data, kernel, *, n_iter=600, recon_type="smf", n_coils=1):
    r"""
    Fit fODF and GM/CSF volume fractions for a voxel using RUMBA-SD.

    Deconvolves the kernel from the diffusion-weighted signal by computing a
    maximum likelihood estimation of the fODF
    :footcite:p:`CanalesRodriguez2015`. Minimizes the negative log-likelihood of
    the data under Rician or Noncentral Chi noise distributions by adapting the
    iterative technique developed in Richardson-Lucy deconvolution.

    Parameters
    ----------
    data : 1d ndarray (N,)
        Signal values for a single voxel.
    kernel : 2d ndarray (N, M)
        Deconvolution kernel mapping volume fractions of the M compartments to
        N-length signal. Last two columns should be for GM and CSF.
    n_iter : int, optional
        Number of iterations for fODF estimation. Must be a positive int.
        Default: 600
    recon_type : {'smf', 'sos'}, optional
        MRI reconstruction method: spatial matched filter (SMF) or
        sum-of-squares (SoS). SMF reconstruction generates Rician noise while
        SoS reconstruction generates Noncentral Chi noise. Default: 'smf'
    n_coils : int, optional
        Number of coils in MRI scanner -- only relevant in SoS reconstruction.
        Must be a positive int. Default: 1

    Returns
    -------
    fit_vec : 1d ndarray (M,)
        Vector containing fODF and GM/CSF volume fractions. First M-2
        components are fODF while last two are GM and CSF respectively.

    Notes
    -----
    The diffusion MRI signal measured at a given voxel is a sum of
    contributions from each intra-voxel compartment, including parallel white
    matter (WM) fiber populations in a given orientation as well as effects
    from GM and CSF. The equation governing these  contributions is:

    $S_i = S_0\left(\sum_{j=1}^{M}f_j\exp(-b_i\textbf{v}_i^T\textbf{D}_j
    \textbf{v}_i) + f_{GM}\exp(-b_iD_{GM})+f_{CSF}\exp(-b_iD_{CSF})\right)$

    Where $S_i$ is the resultant signal along the diffusion-sensitizing
    gradient unit vector $\textbf{v_i}; i = 1, ..., N$ with a b-value of $b_i$.
    $f_j; j = 1, ..., M$ is the volume fraction of the $j^{th}$ fiber
    population with an anisotropic diffusion tensor $\textbf{D_j}$.

    $f_{GM}$ and $f_{CSF}$ are the volume fractions and $D_{GM}$ and $D_{CSF}$
    are the mean diffusivities of GM and CSF respectively.

    This equation is linear in $f_j, f_{GM}, f_{CSF}$ and can be simplified to
    a single matrix multiplication:

    $\textbf{S} = \textbf{Hf}$

    Where $\textbf{S}$ is the signal vector at a certain voxel, $\textbf{H}$ is
    the deconvolution kernel, and $\textbf{f}$ is the vector of volume
    fractions for each compartment.

    Modern MRI scanners produce noise following a Rician or Noncentral Chi
    distribution, depending on their signal reconstruction technique
    `footcite:p:`Constantinides1997`. Using this linear model, it can be shown
    that the likelihood of a signal under a Noncentral Chi noise model is:

    $P(\textbf{S}|\textbf{H}, \textbf{f}, \sigma^2, n) = \prod_{i=1}^{N}\left(
    \frac{S_i}{\bar{S_i}}\right)^n\exp\left\{-\frac{1}{2\sigma^2}\left[
    S_i^2 + \bar{S}_i^2\right]\right\}I_{n-1}\left(\frac{S_i\bar{S}_i}
    {\sigma^2}\right)u(S_i)$

    Where $S_i$ and $\bar{S}_i = \textbf{Hf}$ are the measured and expected
    signals respectively, and $n$ is the number of coils in the scanner, and
    $I_{n-1}$ is the modified Bessel function of first kind of order $n-1$.
    This gives the likelihood under a Rician distribution when $n$ is set to 1.

    By taking the negative log of this with respect to $\textbf{f}$ and setting
    the derivative to 0, the $\textbf{f}$ maximizing likelihood is found to be:

    $\textbf{f} = \textbf{f} \circ \frac{\textbf{H}^T\left[\textbf{S}\circ
    \frac{I_n(\textbf{S}\circ \textbf{Hf}/\sigma^2)} {I_{n-1}(\textbf{S}
    \circ\textbf{Hf}\sigma^2)} \right ]} {\textbf{H}^T\textbf{Hf}}$

    The solution can be found using an iterative scheme, just as in the
    Richardson-Lucy algorithm:

    $\textbf{f}^{k+1} = \textbf{f}^k \circ \frac{\textbf{H}^T\left[\textbf{S}
    \circ\frac{I_n(\textbf{S}\circ\textbf{Hf}^k/\sigma^2)} {I_{n-1}(\textbf{S}
    \circ\textbf{Hf}^k/\sigma^2)} \right ]} {\textbf{H}^T\textbf{Hf}^k}$

    In order to apply this, a reasonable estimate of $\sigma^2$ is required.
    To find this, a separate iterative scheme is found using the derivative
    of the negative log with respect to $\sigma^2$, and is run in parallel.
    This is shown here:

    $\alpha^{k+1} = \frac{1}{nN}\left\{ \frac{\textbf{S}^T\textbf{S} +
    \textbf{f}^T\textbf{H}^T\textbf{Hf}}{2} - \textbf{1}^T_N\left[(\textbf{S}
    \circ\textbf{Hf})\circ\frac{I_n(\textbf{S}\circ\textbf{Hf}/\alpha^k)}
    {I_{n-1}(\textbf{S}\circ\textbf{Hf}/\alpha^k)} \right ]\right \}$

    For more details, see :footcite:p:`CanalesRodriguez2015`.

    References
    ----------
    .. footbibliography::
    """

    n_comp = kernel.shape[1]  # number of compartments
    n_grad = len(data)  # gradient directions

    fodf = np.ones((n_comp, 1))  # initial guess is iso-probable
    fodf = fodf / np.sum(fodf, axis=0)  # normalize initial guess

    if recon_type == "smf":
        n_order = 1  # Rician noise (same as Noncentral Chi with order 1)
    elif recon_type == "sos":
        n_order = n_coils  # Noncentral Chi noise (order = # of coils)
    else:
        raise ValueError(
            f"Invalid recon_type. Should be 'smf' or 'sos', received {recon_type}"
        )

    data = data.reshape(-1, 1)
    reblurred = np.matmul(kernel, fodf)

    # For use later
    kernel_t = kernel.T
    f_zero = 0  # minimum value allowed in fODF

    # Initialize variance
    sigma0 = 1 / 15
    sigma2 = sigma0**2 * np.ones(data.shape)  # Expand into vector

    reblurred_s = data * reblurred / sigma2

    for _ in range(n_iter):
        fodf_i = fodf
        ratio = mbessel_ratio(n_order, reblurred_s)
        rl_factor = np.matmul(kernel_t, data * ratio) / (
            np.matmul(kernel_t, reblurred) + _EPS
        )

        fodf = fodf_i * rl_factor  # result of iteration
        fodf = np.maximum(f_zero, fodf)  # apply positivity constraint

        # Update other variables
        reblurred = np.matmul(kernel, fodf)
        reblurred_s = data * reblurred / sigma2

        # Iterate variance
        sigma2_i = (1 / (n_grad * n_order)) * np.sum(
            (data**2 + reblurred**2) / 2 - (sigma2 * reblurred_s) * ratio, axis=0
        )
        sigma2_i = np.minimum((1 / 8) ** 2, np.maximum(sigma2_i, (1 / 80) ** 2))
        sigma2 = sigma2_i * np.ones(data.shape)  # Expand into vector

    # Normalize final result
    fit_vec = np.squeeze(fodf / (np.sum(fodf, axis=0) + _EPS))

    return fit_vec


def mbessel_ratio(n, x):
    r"""
    Fast computation of modified Bessel function ratio (first kind).

    Computes:

    $I_{n}(x) / I_{n-1}(x)$

    using Perron's continued fraction equation where $I_n$ is the modified
    Bessel function of first kind, order $n$ :footcite:p:`Gautschi1978`.

    Parameters
    ----------
    n : int
        Order of Bessel function in numerator (denominator is of order n-1).
        Must be a positive int.
    x : float or ndarray
        Value or array of values with which to compute ratio.

    Returns
    -------
    y : float or ndarray
        Result of ratio computation.

    References
    ----------
    .. footbibliography::
    """

    y = x / (
        (2 * n + x)
        - (
            2
            * x
            * (n + 1 / 2)
            / (
                2 * n
                + 1
                + 2 * x
                - (
                    2
                    * x
                    * (n + 3 / 2)
                    / (2 * n + 2 + 2 * x - (2 * x * (n + 5 / 2) / (2 * n + 3 + 2 * x)))
                )
            )
        )
    )

    return y


def generate_kernel(gtab, sphere, wm_response, gm_response, csf_response):
    """
    Generate deconvolution kernel

    Compute kernel mapping orientation densities of white matter fiber
    populations (along each vertex of the sphere) and isotropic volume
    fractions to a diffusion weighted signal.

    Parameters
    ----------
    gtab : GradientTable
        Gradient table.
    sphere : Sphere
        Sphere with which to sample discrete fiber orientations in order to
        construct kernel
    wm_response : 1d ndarray or 2d ndarray or AxSymShResponse, optional
        Tensor eigenvalues as a (3,) ndarray, multishell eigenvalues as
        a (len(unique_bvals_tolerance(gtab.bvals))-1, 3) ndarray in
        order of smallest to largest b-value, or an AxSymShResponse.
    gm_response : float, optional
        Mean diffusivity for GM compartment. If `None`, then grey
        matter compartment set to all zeros.
    csf_response : float, optional
        Mean diffusivity for CSF compartment. If `None`, then CSF
        compartment set to all zeros.

    Returns
    -------
    kernel : 2d ndarray (N, M)
        Computed kernel; can be multiplied with a vector consisting of volume
        fractions for each of M-2 fiber populations as well as GM and CSF
        fractions to produce a diffusion weighted signal.
    """

    # Coordinates of sphere vertices
    sticks = sphere.vertices

    n_grad = len(gtab.gradients)  # number of gradient directions
    n_wm_comp = sticks.shape[0]  # number of fiber populations
    n_comp = n_wm_comp + 2  # plus isotropic compartments

    kernel = np.zeros((n_grad, n_comp))

    # White matter compartments
    list_bvals = unique_bvals_tolerance(gtab.bvals)
    n_bvals = len(list_bvals) - 1  # number of unique b-values

    if isinstance(wm_response, AxSymShResponse):
        # Data-driven response
        where_dwi = lazy_index(~gtab.b0s_mask)
        gradients = gtab.gradients[where_dwi]
        gradients = gradients / np.linalg.norm(gradients, axis=1)[..., None]
        S0 = wm_response.S0
        for i in range(n_wm_comp):
            # Response oriented along [0, 0, 1], so must rotate sticks[i]
            rot_mat = vec2vec_rotmat(sticks[i], np.array([0, 0, 1]))
            rot_gradients = np.dot(rot_mat, gradients.T).T
            rot_sphere = Sphere(xyz=rot_gradients)
            # Project onto rotated sphere and scale
            rot_response = wm_response.on_sphere(rot_sphere) / S0
            kernel[where_dwi, i] = rot_response

        # Set b0 components
        kernel[gtab.b0s_mask, :] = 1

    elif wm_response.shape == (n_bvals, 3):
        # Multi-shell response
        bvals = gtab.bvals
        bvecs = gtab.bvecs
        for n, bval in enumerate(list_bvals[1:]):
            indices = get_bval_indices(bvals, bval)
            with warnings.catch_warnings():  # extract relevant b-value
                warnings.simplefilter("ignore")
                gtab_sub = gradient_table(bvals[indices], bvecs=bvecs[indices])

            for i in range(n_wm_comp):
                # Signal generated by WM-fiber for each gradient direction
                S = single_tensor(
                    gtab_sub, evals=wm_response[n], evecs=all_tensor_evecs(sticks[i])
                )
                kernel[indices, i] = S

        # Set b0 components
        b0_indices = get_bval_indices(bvals, list_bvals[0])
        kernel[b0_indices, :] = 1

    else:
        # Single-shell response
        for i in range(n_wm_comp):
            # Signal generated by WM-fiber for each gradient direction
            S = single_tensor(
                gtab, evals=wm_response, evecs=all_tensor_evecs(sticks[i])
            )
            kernel[:, i] = S

        # Set b0 components
        kernel[gtab.b0s_mask, :] = 1

    # GM compartment
    if gm_response is None:
        S_gm = np.zeros(n_grad)
    else:
        S_gm = single_tensor(
            gtab, evals=np.array([gm_response, gm_response, gm_response])
        )

    if csf_response is None:
        S_csf = np.zeros(n_grad)
    else:
        S_csf = single_tensor(
            gtab, evals=np.array([csf_response, csf_response, csf_response])
        )

    kernel[:, n_comp - 2] = S_gm
    kernel[:, n_comp - 1] = S_csf

    return kernel


@warning_for_keywords()
def rumba_deconv_global(
    data,
    kernel,
    mask,
    *,
    n_iter=600,
    recon_type="smf",
    n_coils=1,
    R=1,
    use_tv=True,
    verbose=False,
):
    r"""
    Fit fODF for all voxels simultaneously using RUMBA-SD.

    Deconvolves the kernel from the diffusion-weighted signal at each voxel by
    computing a maximum likelihood estimation of the fODF
    :footcite:p:`CanalesRodriguez2015`. Global fitting also permits the use of
    total variation regularization (RUMBA-SD + TV). The spatial dependence
    introduced by TV promotes smoother solutions (i.e. prevents oscillations),
    while still allowing for sharp discontinuities :footcite:p:`Rudin1992`. This
    promotes smoothness and continuity along individual tracts while preventing
    smoothing of adjacent tracts.

    Generally, global_fit will proceed more quickly than the voxelwise fit
    provided that the computer has adequate RAM (>= 16 GB should be more than
    sufficient).

    Parameters
    ----------
    data : 4d ndarray (x, y, z, N)
        Signal values for entire brain. None of the volume dimensions x, y, z
        can be 1 if TV regularization is required.
    kernel : 2d ndarray (N, M)
        Deconvolution kernel mapping volume fractions of the M compartments to
        N-length signal. Last two columns should be for GM and CSF.
    mask : 3d ndarray(x, y, z)
        Binary mask specifying voxels of interest with 1; fODF will only be
        fit at these voxels (0 elsewhere).
    n_iter : int, optional
        Number of iterations for fODF estimation. Must be a positive int.
    recon_type : {'smf', 'sos'}, optional
        MRI reconstruction method: spatial matched filter (SMF) or
        sum-of-squares (SoS). SMF reconstruction generates Rician noise while
        SoS reconstruction generates Noncentral Chi noise.
    n_coils : int, optional
        Number of coils in MRI scanner -- only relevant in SoS reconstruction.
        Must be a positive int.
    use_tv : bool, optional
        If true, applies total variation regularization. This requires a brain
        volume with no singleton dimensions.
    verbose : bool, optional
        If true, logs updates on estimated signal-to-noise ratio after each
        iteration.

    Returns
    -------
    fit_array : 4d ndarray (x, y, z, M)
        fODF and GM/CSF volume fractions computed for each voxel. First M-2
        components are fODF, while last two are GM and CSf respectively.

    Notes
    -----
    TV modifies our cost function as follows:

    .. math::

    J(\textbf{f}) = -\log{P(\textbf{S}|\textbf{H}, \textbf{f}, \sigma^2, n)})+
    \alpha_{TV}TV(\textbf{f})

    where the first term is the negative log likelihood described in the notes
    of `rumba_deconv`, and the second term is the TV energy, or the sum of
    gradient absolute values for the fODF across the entire brain. This results
    in a new multiplicative factor in the iterative scheme, now becoming:

    .. math::

    \textbf{f}^{k+1} = \textbf{f}^k \circ \frac{\textbf{H}^T\left[\textbf{S}
    \circ\frac{I_n(\textbf{S}\circ\textbf{Hf}^k/\sigma^2)} {I_{n-1}(\textbf{S}
    \circ\textbf{Hf}^k/\sigma^2)} \right ]} {\textbf{H}^T\textbf{Hf}^k}\circ
    \textbf{R}^k

    where $\textbf{R}^k$ is computed voxelwise by:

    .. math::

    (\textbf{R}^k)_j = \frac{1}{1 - \alpha_{TV}div\left(\frac{\triangledown[
    \textbf{f}^k_{3D}]_j}{\lvert\triangledown[\textbf{f}^k_{3D}]_j \rvert}
    \right)\biggr\rvert_{x, y, z}}

    Here, $\triangledown$ is the symbol for the 3D gradient at any voxel.

    The regularization strength, $\alpha_{TV}$ is updated after each iteration
    by the discrepancy principle -- specifically, it is selected to match the
    estimated variance after each iteration :footcite:p:`Chambolle2004`.

    References
    ----------
    .. footbibliography::
    """

    # Crop data to reduce memory consumption
    dim_orig = data.shape
    ixmin, ixmax = bounding_box(mask)
    data = crop(data, ixmin, ixmax)
    mask = crop(mask, ixmin, ixmax)

    if np.any(np.array(data.shape[:3]) == 1) and use_tv:
        raise ValueError(
            "Cannot use TV regularization if any spatial"
            + "dimensions are 1; "
            + f"provided dimensions were {data.shape[:3]}"
        )

    epsilon = 1e-7

    n_grad = kernel.shape[0]  # gradient directions
    n_comp = kernel.shape[1]  # number of compartments
    dim = data.shape
    n_v_tot = np.prod(dim[:3])  # total number of voxels

    # Initial guess is iso-probable
    fodf0 = np.ones((n_comp, 1), dtype=np.float32)
    fodf0 = fodf0 / np.sum(fodf0, axis=0)

    if recon_type == "smf":
        n_order = 1  # Rician noise (same as Noncentral Chi with order 1)
    elif recon_type == "sos":
        n_order = n_coils  # Noncentral Chi noise (order = # of coils)
    else:
        raise ValueError(
            f"Invalid recon_type. Should be 'smf' or 'sos', received f{recon_type}"
        )

    mask_vec = np.ravel(mask)
    # Indices of target voxels
    index_mask = np.atleast_1d(np.squeeze(np.argwhere(mask_vec)))
    n_v_true = len(index_mask)  # number of target voxels

    data_2d = np.zeros((n_v_true, n_grad), dtype=np.float32)
    for i in range(n_grad):
        data_2d[:, i] = np.ravel(data[:, :, :, i])[
            index_mask
        ]  # only keep voxels of interest

    data_2d = data_2d.T
    fodf = np.tile(fodf0, (1, n_v_true))
    reblurred = np.matmul(kernel, fodf)

    # For use later
    kernel_t = kernel.T
    f_zero = 0

    # Initialize algorithm parameters
    sigma0 = 1 / 15
    sigma2 = sigma0**2
    tv_lambda = sigma2  # initial guess for TV regularization strength

    # Expand into matrix form for iterations
    sigma2 = sigma2 * np.ones(data_2d.shape, dtype=np.float32)
    tv_lambda_aux = np.zeros(n_v_tot, dtype=np.float32)

    reblurred_s = data_2d * reblurred / sigma2

    for i in range(n_iter):
        fodf_i = fodf
        ratio = mbessel_ratio(n_order, reblurred_s).astype(np.float32)
        rl_factor = np.matmul(kernel_t, data_2d * ratio) / (
            np.matmul(kernel_t, reblurred) + _EPS
        )

        if use_tv:  # apply TV regularization
            tv_factor = np.ones(fodf_i.shape, dtype=np.float32)
            fodf_4d = _reshape_2d_4d(fodf_i.T, mask)
            # Compute gradient, divergence
            gr = _grad(fodf_4d)
            d_inv = 1 / np.sqrt(epsilon**2 + np.sum(gr**2, axis=3))
            gr_norm = gr * d_inv[:, :, :, None, :]
            div_f = _divergence(gr_norm)
            g0 = np.abs(1 - tv_lambda * div_f)
            tv_factor_4d = 1 / (g0 + _EPS)

            for j in range(n_comp):
                tv_factor_1d = np.ravel(tv_factor_4d[:, :, :, j])[index_mask]
                tv_factor[j, :] = tv_factor_1d

            # Apply TV regularization to iteration factor
            rl_factor = rl_factor * tv_factor

        fodf = fodf_i * rl_factor  # result of iteration
        fodf = np.maximum(f_zero, fodf)  # positivity constraint

        # Update other variables
        reblurred = np.matmul(kernel, fodf)
        reblurred_s = data_2d * reblurred / sigma2

        # Iterate variance
        sigma2_i = (1 / (n_grad * n_order)) * np.sum(
            (data_2d**2 + reblurred**2) / 2 - (sigma2 * reblurred_s) * ratio, axis=0
        )
        sigma2_i = np.minimum((1 / 8) ** 2, np.maximum(sigma2_i, (1 / 80) ** 2))

        if verbose:
            logger.info("Iteration %d of %d", i + 1, n_iter)

            snr_mean = np.mean(1 / np.sqrt(sigma2_i))
            snr_std = np.std(1 / np.sqrt(sigma2_i))
            logger.info(
                "Mean SNR (S0/sigma) estimated to be %.3f +/- %.3f", snr_mean, snr_std
            )
        # Expand into matrix
        sigma2 = np.tile(sigma2_i[None, :], (data_2d.shape[0], 1))

        # Update TV regularization strength using the discrepancy principle
        if use_tv:
            if R == 1:
                tv_lambda = np.mean(sigma2_i)

                if tv_lambda < (1 / 30) ** 2:
                    tv_lambda = (1 / 30) ** 2
            else:  # different factor for each voxel
                tv_lambda_aux[index_mask] = sigma2_i
                tv_lambda = np.reshape(tv_lambda_aux, (*dim[:3], 1))

    fodf = fodf.astype(np.float64)
    fodf = fodf / (np.sum(fodf, axis=0)[None, ...] + _EPS)  # normalize fODF

    # Extract compartments
    fit_array = np.zeros((*dim_orig[:3], n_comp))
    _reshape_2d_4d(
        fodf.T,
        mask,
        out=fit_array[ixmin[0] : ixmax[0], ixmin[1] : ixmax[1], ixmin[2] : ixmax[2]],
    )

    return fit_array


def _grad(M):
    """
    Computes one way first difference
    """
    x_ind = list(range(1, M.shape[0])) + [M.shape[0] - 1]
    y_ind = list(range(1, M.shape[1])) + [M.shape[1] - 1]
    z_ind = list(range(1, M.shape[2])) + [M.shape[2] - 1]

    grad = np.zeros((*M.shape[:3], 3, M.shape[-1]), dtype=np.float32)
    grad[:, :, :, 0, :] = M[x_ind, :, :, :] - M
    grad[:, :, :, 1, :] = M[:, y_ind, :, :] - M
    grad[:, :, :, 2, :] = M[:, :, z_ind, :] - M

    return grad


def _divergence(F):
    """
    Computes divergence of a 3-dimensional vector field (with one way
    first difference)
    """
    Fx = F[:, :, :, 0, :]
    Fy = F[:, :, :, 1, :]
    Fz = F[:, :, :, 2, :]

    x_ind = [0] + list(range(F.shape[0] - 1))
    y_ind = [0] + list(range(F.shape[1] - 1))
    z_ind = [0] + list(range(F.shape[2] - 1))

    fx = Fx - Fx[x_ind, :, :, :]
    fx[0, :, :, :] = Fx[0, :, :, :]  # edge conditions
    fx[-1, :, :, :] = -Fx[-2, :, :, :]

    fy = Fy - Fy[:, y_ind, :, :]
    fy[:, 0, :, :] = Fy[:, 0, :, :]
    fy[:, -1, :, :] = -Fy[:, -2, :, :]

    fz = Fz - Fz[:, :, z_ind, :]
    fz[:, :, 0, :] = Fz[:, :, 0, :]
    fz[:, :, -1, :] = -Fz[:, :, -2, :]

    return fx + fy + fz


@warning_for_keywords()
def _reshape_2d_4d(M, mask, *, out=None):
    """
    Faster reshape from 2D to 4D.
    """
    if out is None:
        out = np.zeros((*mask.shape, M.shape[-1]), dtype=M.dtype)
    n = 0
    for i, j, k in np.ndindex(mask.shape):
        if mask[i, j, k]:
            out[i, j, k, :] = M[n, :]
            n += 1
    return out



import warnings

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

plt, have_plt, _ = optional_package("matplotlib.pyplot")
fury, has_fury, _ = optional_package("fury", min_version="0.10.0")
if has_fury:
    from dipy.viz import actor, window

sagittal_deprecation_warning_msg = (
    "The view argument value `sagital` is deprecated and "  # codespell:ignore sagital
    "its support will be removed in a future version. Please, use `sagittal` instead."
)


@warning_for_keywords()
def show_bundles(
    bundles,
    *,
    interactive=True,
    view="sagittal",
    colors=None,
    linewidth=0.3,
    save_as=None,
):
    """Render bundles to visualize them interactively or save them into a png.

    The function allows to just render the bundles in an interactive plot or
    to export them into a png file.

    Parameters
    ----------
    bundles : list
        Bundles to be rendered.
    interactive : boolean, optional
        If True a 3D interactive rendering is created. Default is True.
    view : str, optional
        Viewing angle. Supported options: 'sagittal', 'axial' and 'coronal'.
    colors : list, optional
       Colors to be used for each bundle. If None default colors are used.
    linewidth : float, optional
        Width of each rendered streamline. Default is 0.3.
    save_as : str, optional
        If not None rendered scene is stored in a png file with that name.
        Default is None.

    """

    if view == "sagital":  # codespell:ignore sagital
        warnings.warn(
            sagittal_deprecation_warning_msg,
            category=DeprecationWarning,
            stacklevel=2,
        )
        view = "sagittal"

    scene = window.Scene()
    scene.SetBackground(1.0, 1, 1)

    for i, bundle in enumerate(bundles):
        if colors is None:
            lines_actor = actor.streamtube(bundle, linewidth=linewidth)
        else:
            lines_actor = actor.streamtube(
                bundle, linewidth=linewidth, colors=colors[i]
            )
        if view == "sagittal":
            lines_actor.RotateX(-90)
            lines_actor.RotateZ(90)
        elif view == "axial":
            pass
        elif view == "coronal":
            lines_actor.RotateX(-90)
        else:
            raise ValueError(
                "Invalid view argument value. Use 'sagittal', 'axial' or 'coronal'."
            )

        scene.add(lines_actor)

    if interactive:
        window.show(scene)

    if save_as is not None:
        window.record(scene=scene, n_frames=1, out_path=save_as, size=(900, 900))


@warning_for_keywords()
def viz_two_bundles(b1, b2, fname, *, c1=(1, 0, 0), c2=(0, 1, 0), interactive=False):
    """Render and plot two bundles to visualize them.

    Parameters
    ----------
    b1 : Streamlines
        Bundle one to be rendered.
    b2 : Streamlines
        Bundle two to be rendered.
    fname: str
        Rendered scene is stored in a png file with that name.
    C1 : tuple, optional
        Color to be used for first bundle. Default red.
    C2 : tuple, optional
        Color to be used for second bundle. Default green.
    interactive : boolean, optional
        If True a 3D interactive rendering is created. Default is True.

    """
    ren = window.Scene()
    ren.SetBackground(1, 1, 1)

    actor1 = actor.line(b1, colors=c1)
    actor1.GetProperty().SetEdgeVisibility(1)
    actor1.GetProperty().SetRenderLinesAsTubes(1)
    actor1.GetProperty().SetLineWidth(6)
    actor1.GetProperty().SetOpacity(1)
    actor1.RotateX(-70)
    actor1.RotateZ(90)

    ren.add(actor1)

    actor2 = actor.line(b2, colors=c2)
    actor2.GetProperty().SetEdgeVisibility(1)
    actor2.GetProperty().SetRenderLinesAsTubes(1)
    actor2.GetProperty().SetLineWidth(6)
    actor2.GetProperty().SetOpacity(1)
    actor2.RotateX(-70)
    actor2.RotateZ(90)

    ren.add(actor2)

    if interactive:
        window.show(ren)

    window.record(scene=ren, n_frames=1, out_path=fname, size=(1200, 1200))
    if interactive:
        im = plt.imread(fname)
        plt.figure(figsize=(10, 10))
        plt.imshow(im)


@warning_for_keywords()
def viz_vector_field(
    points_aligned,
    directions,
    colors,
    offsets,
    fname,
    *,
    bundle=None,
    interactive=False,
):
    """Render and plot vector field.

    Parameters
    ----------
    points_aligned : List
        List containing starting positions of vectors.
    directions : List
        List containing unitary directions of vectors.
    colors : List
        List containing colors for each vector.
    offsets : List
        List containing vector field modules.
    fname: str
        Rendered scene is stored in a png file with that name.
    bundle : Streamlines, optional
        Bundle to be rendered with vector field (Default None).
    interactive : boolean, optional
        If True a 3D interactive rendering is created. Default is True.

    """
    scene = window.Scene()
    scene.SetBackground(1.0, 1, 1)
    arrows = actor.arrow(points_aligned, directions, colors, scales=offsets)
    arrows.RotateX(-70)
    arrows.RotateZ(90)
    scene.add(arrows)

    if bundle:
        actor1 = actor.line(bundle, colors=(0, 0, 1))
        actor1.RotateX(-70)
        actor1.RotateZ(90)
        scene.add(actor1)

    if interactive:
        window.show(scene)

    window.record(scene=scene, n_frames=1, out_path=fname, size=(1200, 1200))
    if interactive:
        im = plt.imread(fname)
        plt.figure(figsize=(10, 10))
        plt.imshow(im)


@warning_for_keywords()
def viz_displacement_mag(bundle, offsets, fname, *, interactive=False):
    """Render and plot displacement magnitude over the bundle.

    Parameters
    ----------
    bundle : Streamlines,
        Bundle to be rendered.
    offsets : List
        List containing displacement magnitdues per point on the bundle.
    fname: str
        Rendered scene is stored in a png file with that name.
    interactive : boolean, optional
        If True a 3D interactive rendering is created. Default is True.

    """
    scene = window.Scene()
    hue = (0.1, 0.9)
    hue = (0.9, 0.3)
    saturation = (0.5, 1)
    scene.background((1, 1, 1))
    lut_cmap = actor.colormap_lookup_table(
        scale_range=(offsets.min(), offsets.max()),
        hue_range=hue,
        saturation_range=saturation,
    )

    stream_actor = actor.line(
        bundle, colors=offsets, linewidth=7, lookup_colormap=lut_cmap
    )

    stream_actor.RotateX(-70)
    stream_actor.RotateZ(90)

    scene.add(stream_actor)
    bar = actor.scalar_bar(lookup_table=lut_cmap)

    scene.add(bar)

    if interactive:
        window.show(scene)

    window.record(scene=scene, n_frames=1, out_path=fname, size=(2000, 1500))
    if interactive:
        im = plt.imread(fname)
        plt.figure(figsize=(10, 10))
        plt.imshow(im)



from abc import ABCMeta, abstractmethod
import logging
import operator
from time import time

import numpy as np

from dipy.segment.featurespeed import ResampleFeature
from dipy.segment.metricspeed import (
    AveragePointwiseEuclideanMetric,
    Metric,
    MinimumAverageDirectFlipMetric,
)
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import nbytes, set_number_of_points

logger = logging.getLogger(__name__)


class Identity:
    """Provides identity indexing functionality.

    This can replace any class supporting indexing used for referencing
    (e.g. list, tuple). Indexing an instance of this class will return the
    index provided instead of the element. It does not support slicing.
    """

    def __getitem__(self, idx):
        return idx


class Cluster:
    """Provides functionalities for interacting with a cluster.

    Useful container to retrieve index of elements grouped together. If
    a reference to the data is provided to `cluster_map`, elements will
    be returned instead of their index when possible.

    Parameters
    ----------
    cluster_map : `ClusterMap` object
        Reference to the set of clusters this cluster is being part of.
    id : int, optional
        Id of this cluster in its associated `cluster_map` object.
    refdata : list, optional
        Actual elements that clustered indices refer to.

    Notes
    -----
    A cluster does not contain actual data but instead knows how to
    retrieve them using its `ClusterMap` object.
    """

    @warning_for_keywords()
    def __init__(self, *, id=0, indices=None, refdata=None):
        if refdata is None:
            refdata = Identity()
        self.id = id
        self.refdata = refdata
        self.indices = indices if indices is not None else []

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, idx):
        """Gets element(s) through indexing.

        If a reference to the data was provided (via refdata property)
        elements will be returned instead of their index.

        Parameters
        ----------
        idx : int, slice or list
            Index of the element(s) to get.

        Returns
        -------
        `Cluster` object(s)
            When `idx` is a int, returns a single element.

            When `idx` is either a slice or a list, returns a list of elements.
        """
        if isinstance(idx, (int, np.integer)):
            return self.refdata[self.indices[idx]]
        elif isinstance(idx, slice):
            return [self.refdata[i] for i in self.indices[idx]]
        elif isinstance(idx, list):
            return [self[i] for i in idx]

        msg = f"Index must be a int or a slice! Not '{type(idx)}'"
        raise TypeError(msg)

    def __iter__(self):
        return (self[i] for i in range(len(self)))

    def __str__(self):
        return "[" + ", ".join(map(str, self.indices)) + "]"

    def __repr__(self):
        return f"Cluster({str(self)})"

    def __eq__(self, other):
        return isinstance(other, Cluster) and self.indices == other.indices

    def __ne__(self, other):
        return not self == other

    def __cmp__(self, other):
        raise TypeError("Cannot compare Cluster objects.")

    def assign(self, *indices):
        """Assigns indices to this cluster.

        Parameters
        ----------
        *indices : list of indices
            Indices to add to this cluster.
        """
        self.indices += indices


class ClusterCentroid(Cluster):
    """Provides functionalities for interacting with a cluster.

    Useful container to retrieve the indices of elements grouped together and
    the cluster's centroid. If a reference to the data is provided to
    `cluster_map`, elements will be returned instead of their index when
    possible.

    Parameters
    ----------
    cluster_map : `ClusterMapCentroid` object
        Reference to the set of clusters this cluster is being part of.
    id : int, optional
        Id of this cluster in its associated `cluster_map` object.
    refdata : list, optional
        Actual elements that clustered indices refer to.

    Notes
    -----
    A cluster does not contain actual data but instead knows how to
    retrieve them using its `ClusterMapCentroid` object.
    """

    @warning_for_keywords()
    def __init__(self, centroid, *, id=0, indices=None, refdata=None):
        if refdata is None:
            refdata = Identity()
        super(ClusterCentroid, self).__init__(id=id, indices=indices, refdata=refdata)
        self.centroid = centroid.copy()
        self.new_centroid = centroid.copy()

    def __eq__(self, other):
        return (
            isinstance(other, ClusterCentroid)
            and np.all(self.centroid == other.centroid)
            and super(ClusterCentroid, self).__eq__(other)
        )

    def assign(self, id_datum, features):
        """Assigns a data point to this cluster.

        Parameters
        ----------
        id_datum : int
            Index of the data point to add to this cluster.
        features : 2D array
            Data point's features to modify this cluster's centroid.
        """
        N = len(self)
        self.new_centroid = ((self.new_centroid * N) + features) / (N + 1.0)
        super(ClusterCentroid, self).assign(id_datum)

    def update(self):
        """Update centroid of this cluster.

        Returns
        -------
        converged : bool
            Tells if the centroid has moved.
        """
        converged = np.equal(self.centroid, self.new_centroid)
        self.centroid = self.new_centroid.copy()
        return converged


class ClusterMap:
    """Provides functionalities for interacting with clustering outputs.

    Useful container to create, remove, retrieve and filter clusters.
    If `refdata` is given, elements will be returned instead of their
    index when using `Cluster` objects.

    Parameters
    ----------
    refdata : list
        Actual elements that clustered indices refer to.
    """

    @warning_for_keywords()
    def __init__(self, *, refdata=None):
        if refdata is None:
            refdata = Identity()
        self._clusters = []
        self.refdata = refdata

    @property
    def clusters(self):
        return self._clusters

    @property
    def refdata(self):
        return self._refdata

    @refdata.setter
    def refdata(self, value):
        if value is None:
            value = Identity()

        self._refdata = value
        for cluster in self.clusters:
            cluster.refdata = self._refdata

    def __len__(self):
        return len(self.clusters)

    def __getitem__(self, idx):
        """Gets cluster(s) through indexing.

        Parameters
        ----------
        idx : int, slice, list or boolean array
            Index of the element(s) to get.

        Returns
        -------
        `Cluster` object(s)
            When `idx` is an int, returns a single `Cluster` object.

            When `idx`is either a slice, list or boolean array, returns
            a list of `Cluster` objects.
        """
        if isinstance(idx, np.ndarray) and idx.dtype == bool:
            return [self.clusters[i] for i, take_it in enumerate(idx) if take_it]
        elif isinstance(idx, slice):
            return [self.clusters[i] for i in range(*idx.indices(len(self)))]
        elif isinstance(idx, list):
            return [self.clusters[i] for i in idx]

        return self.clusters[idx]

    def __iter__(self):
        return iter(self.clusters)

    def __str__(self):
        return "[" + ", ".join(map(str, self)) + "]"

    def __repr__(self):
        return f"ClusterMap({str(self)})"

    def _richcmp(self, other, op):
        """Compares this cluster map with another cluster map or an integer.

        Two `ClusterMap` objects are equal if they contain the same clusters.
        When comparing a `ClusterMap` object with an integer, the comparison
        will be performed on the size of the clusters instead.

        Parameters
        ----------
        other : `ClusterMap` object or int
            Object to compare to.
        op : rich comparison operators (see module `operator`)
            Valid operators are: lt, le, eq, ne, gt or ge.

        Returns
        -------
        bool or 1D array (bool)
            When comparing to another `ClusterMap` object, it returns whether
            the two `ClusterMap` objects contain the same clusters or not.

            When comparing to an integer the comparison is performed on the
            clusters sizes, it returns an array of boolean.
        """
        if isinstance(other, ClusterMap):
            if op is operator.eq:
                return (
                    isinstance(other, ClusterMap)
                    and len(self) == len(other)
                    and self.clusters == other.clusters
                )
            elif op is operator.ne:
                return not self == other

            raise NotImplementedError(
                "Can only check if two ClusterMap instances are equal or not."
            )

        elif isinstance(other, int):
            return np.array([op(len(cluster), other) for cluster in self])

        msg = (
            "ClusterMap only supports comparison with a int or another"
            " instance of Clustermap."
        )
        raise NotImplementedError(msg)

    def __eq__(self, other):
        return self._richcmp(other, operator.eq)

    def __ne__(self, other):
        return self._richcmp(other, operator.ne)

    def __lt__(self, other):
        return self._richcmp(other, operator.lt)

    def __le__(self, other):
        return self._richcmp(other, operator.le)

    def __gt__(self, other):
        return self._richcmp(other, operator.gt)

    def __ge__(self, other):
        return self._richcmp(other, operator.ge)

    def add_cluster(self, *clusters):
        """Adds one or multiple clusters to this cluster map.

        Parameters
        ----------
        *clusters : `Cluster` object, ...
            Cluster(s) to be added in this cluster map.
        """
        for cluster in clusters:
            self.clusters.append(cluster)
            cluster.refdata = self.refdata

    def remove_cluster(self, *clusters):
        """Remove one or multiple clusters from this cluster map.

        Parameters
        ----------
        *clusters : `Cluster` object, ...
            Cluster(s) to be removed from this cluster map.
        """
        for cluster in clusters:
            self.clusters.remove(cluster)

    def clear(self):
        """Remove all clusters from this cluster map."""
        del self.clusters[:]

    def size(self):
        """Gets number of clusters contained in this cluster map."""
        return len(self)

    def clusters_sizes(self):
        """Gets the size of every cluster contained in this cluster map.

        Returns
        -------
        list of int
            Sizes of every cluster in this cluster map.
        """
        return list(map(len, self))

    def get_large_clusters(self, min_size):
        """Gets clusters which contains at least `min_size` elements.

        Parameters
        ----------
        min_size : int
            Minimum number of elements a cluster needs to have to be selected.

        Returns
        -------
        list of `Cluster` objects
            Clusters having at least `min_size` elements.
        """
        return self[self >= min_size]

    def get_small_clusters(self, max_size):
        """Gets clusters which contains at most `max_size` elements.

        Parameters
        ----------
        max_size : int
            Maximum number of elements a cluster can have to be selected.

        Returns
        -------
        list of `Cluster` objects
            Clusters having at most `max_size` elements.
        """
        return self[self <= max_size]


class ClusterMapCentroid(ClusterMap):
    """Provides functionalities for interacting with clustering outputs
    that have centroids.

    Allows to retrieve easily the centroid of every cluster. Also, it is
    a useful container to create, remove, retrieve and filter clusters.
    If `refdata` is given, elements will be returned instead of their
    index when using `ClusterCentroid` objects.

    Parameters
    ----------
    refdata : list
        Actual elements that clustered indices refer to.
    """

    @property
    def centroids(self):
        return [cluster.centroid for cluster in self.clusters]


class Clustering:
    __metaclass__ = ABCMeta

    @abstractmethod
    @warning_for_keywords()
    def cluster(self, data, *, ordering=None):
        """Clusters `data`.

        Subclasses will perform their clustering algorithm here.

        Parameters
        ----------
        data : list of N-dimensional arrays
            Each array represents a data point.
        ordering : iterable of indices, optional
            Specifies the order in which data points will be clustered.

        Returns
        -------
        `ClusterMap` object
            Result of the clustering.
        """
        msg = "Subclass has to define method 'cluster(data, ordering)'!"
        raise NotImplementedError(msg)


class QuickBundles(Clustering):
    r"""Clusters streamlines using QuickBundles.

    Given a list of streamlines, the QuickBundles algorithm
    :footcite:p:`Garyfallidis2012a` sequentially assigns each streamline to its
    closest bundle in $\mathcal{O}(Nk)$ where $N$ is the number of streamlines
    and $k$ is the final number of bundles. If for a given streamline its
    closest bundle is farther than `threshold`, a new bundle is created and the
    streamline is assigned to it except if the number of bundles has already
    exceeded `max_nb_clusters`.

    Parameters
    ----------
    threshold : float
        The maximum distance from a bundle for a streamline to be still
        considered as part of it.
    metric : str or `Metric` object, optional
        The distance metric to use when comparing two streamlines. By default,
        the Minimum average Direct-Flip (MDF) distance
        :footcite:p:`Garyfallidis2012a` is used and streamlines are
        automatically resampled so they have 12 points.
    max_nb_clusters : int, optional
        Limits the creation of bundles.

    Examples
    --------
    >>> from dipy.segment.clustering import QuickBundles
    >>> from dipy.data import get_fnames
    >>> from dipy.io.streamline import load_tractogram
    >>> from dipy.tracking.streamline import Streamlines
    >>> fname = get_fnames(name='fornix')
    >>> fornix = load_tractogram(fname, 'same',
    ...                          bbox_valid_check=False).streamlines
    >>> streamlines = Streamlines(fornix)
    >>> # Segment fornix with a threshold of 10mm and streamlines resampled
    >>> # to 12 points.
    >>> qb = QuickBundles(threshold=10.)
    >>> clusters = qb.cluster(streamlines)
    >>> len(clusters)
    4
    >>> list(map(len, clusters))
    [61, 191, 47, 1]
    >>> # Resampling streamlines differently is done explicitly as follows.
    >>> # Note this has an impact on the speed and the accuracy (tradeoff).
    >>> from dipy.segment.featurespeed import ResampleFeature
    >>> from dipy.segment.metricspeed import AveragePointwiseEuclideanMetric
    >>> feature = ResampleFeature(nb_points=2)
    >>> metric = AveragePointwiseEuclideanMetric(feature)
    >>> qb = QuickBundles(threshold=10., metric=metric)
    >>> clusters = qb.cluster(streamlines)
    >>> len(clusters)
    4
    >>> list(map(len, clusters))
    [58, 142, 72, 28]


    References
    ----------
    .. footbibliography::
    """

    @warning_for_keywords()
    def __init__(self, threshold, *, metric="MDF_12points", max_nb_clusters=None):
        if max_nb_clusters is None:
            max_nb_clusters = np.iinfo("i4").max

        self.threshold = threshold
        self.max_nb_clusters = max_nb_clusters

        if isinstance(metric, MinimumAverageDirectFlipMetric):
            raise ValueError("Use AveragePointwiseEuclideanMetric instead")

        if isinstance(metric, Metric):
            self.metric = metric
        elif metric == "MDF_12points":
            feature = ResampleFeature(nb_points=12)
            self.metric = AveragePointwiseEuclideanMetric(feature)
        else:
            raise ValueError(f"Unknown metric: {metric}")

    @warning_for_keywords()
    def cluster(self, streamlines, *, ordering=None):
        """Clusters `streamlines` into bundles.

        Performs quickbundles algorithm using predefined metric and threshold.

        Parameters
        ----------
        streamlines : list of 2D arrays
            Each 2D array represents a sequence of 3D points (points, 3).
        ordering : iterable of indices, optional
            Specifies the order in which data points will be clustered.

        Returns
        -------
        `ClusterMapCentroid` object
            Result of the clustering.
        """
        from dipy.segment.clustering_algorithms import quickbundles

        cluster_map = quickbundles(
            streamlines,
            self.metric,
            threshold=self.threshold,
            max_nb_clusters=self.max_nb_clusters,
            ordering=ordering,
        )

        cluster_map.refdata = streamlines
        return cluster_map


class QuickBundlesX(Clustering):
    r"""Clusters streamlines using QuickBundlesX.

    See :footcite:p:`Garyfallidis2016` for further details about the method.

    Parameters
    ----------
    thresholds : list of float
        Thresholds to use for each clustering layer. A threshold represents the
        maximum distance from a cluster for a streamline to be still considered
        as part of it.
    metric : str or `Metric` object, optional
        The distance metric to use when comparing two streamlines. By default,
        the Minimum average Direct-Flip (MDF) distance
        :footcite:p:`Garyfallidis2012a` is used and streamlines are
        automatically resampled so they have 12 points.

    References
    ----------
    .. footbibliography::
    """

    @warning_for_keywords()
    def __init__(self, thresholds, *, metric="MDF_12points"):
        self.thresholds = thresholds

        if isinstance(metric, MinimumAverageDirectFlipMetric):
            raise ValueError("Use AveragePointwiseEuclideanMetric instead")

        if isinstance(metric, Metric):
            self.metric = metric
        elif metric == "MDF_12points":
            feature = ResampleFeature(nb_points=12)
            self.metric = AveragePointwiseEuclideanMetric(feature)
        else:
            raise ValueError(f"Unknown metric: {metric}")

    @warning_for_keywords()
    def cluster(self, streamlines, *, ordering=None):
        """Clusters `streamlines` into bundles.

        Performs QuickbundleX using a predefined metric and thresholds.

        Parameters
        ----------
        streamlines : list of 2D arrays
            Each 2D array represents a sequence of 3D points (points, 3).
        ordering : iterable of indices
            Specifies the order in which data points will be clustered.

        Returns
        -------
        `TreeClusterMap` object
            Result of the clustering.
        """
        from dipy.segment.clustering_algorithms import quickbundlesx

        tree = quickbundlesx(
            streamlines, self.metric, thresholds=self.thresholds, ordering=ordering
        )
        tree.refdata = streamlines
        return tree


class TreeCluster(ClusterCentroid):
    @warning_for_keywords()
    def __init__(self, threshold, centroid, *, indices=None):
        super(TreeCluster, self).__init__(centroid=centroid, indices=indices)
        self.threshold = threshold
        self.parent = None
        self.children = []

    def add(self, child):
        child.parent = self
        self.children.append(child)

    @property
    def is_leaf(self):
        return len(self.children) == 0

    def return_indices(self):
        return self.children


class TreeClusterMap(ClusterMap):
    def __init__(self, root):
        self.root = root
        self.leaves = []

        def _retrieves_leaves(node):
            if node.is_leaf:
                self.leaves.append(node)

        self.traverse_postorder(self.root, _retrieves_leaves)

    @property
    def refdata(self):
        return self._refdata

    @refdata.setter
    def refdata(self, value):
        if value is None:
            value = Identity()

        self._refdata = value

        def _set_refdata(node):
            node.refdata = self._refdata

        self.traverse_postorder(self.root, _set_refdata)

    def traverse_postorder(self, node, visit):
        for child in node.children:
            self.traverse_postorder(child, visit)

        visit(node)

    def iter_preorder(self, node):
        parent_stack = []
        while len(parent_stack) > 0 or node is not None:
            if node is not None:
                yield node
                if len(node.children) > 0:
                    parent_stack += node.children[1:]
                    node = node.children[0]
                else:
                    node = None
            else:
                node = parent_stack.pop()

    def __iter__(self):
        return self.iter_preorder(self.root)

    def get_clusters(self, wanted_level):
        clusters = ClusterMapCentroid()

        def _traverse(node, level=0):
            if level == wanted_level:
                clusters.add_cluster(node)
                return

            for child in node.children:
                _traverse(child, level + 1)

        _traverse(self.root)
        return clusters


@warning_for_keywords()
def qbx_and_merge(
    streamlines, thresholds, *, nb_pts=20, select_randomly=None, rng=None, verbose=False
):
    """Run QuickBundlesX and then run again on the centroids of the last layer.

    Running again QuickBundles at a layer has the effect of merging
    some of the clusters that may be originally divided because of branching.
    This function help obtain a result at a QuickBundles quality but with
    QuickBundlesX speed. The merging phase has low cost because it is applied
    only on the centroids rather than the entire dataset.

    See :footcite:p:`Garyfallidis2012a` and :footcite:p:`Garyfallidis2016` for
    further details about the method.

    Parameters
    ----------
    streamlines : Streamlines
        Streamlines.
    thresholds : sequence
        List of distance thresholds for QuickBundlesX.
    nb_pts : int
        Number of points for discretizing each streamline
    select_randomly : int
        Randomly select a specific number of streamlines. If None all the
        streamlines are used.
    rng : numpy.random.Generator
        If None then generator is initialized internally.
    verbose : bool, optional.
        If True, log information. Default False.

    Returns
    -------
    clusters : obj
        Contains the clusters of the last layer of QuickBundlesX after merging.

    References
    ----------
    .. footbibliography::
    """
    t = time()
    len_s = len(streamlines)
    if select_randomly is None:
        select_randomly = len_s

    if rng is None:
        rng = np.random.default_rng()
    indices = rng.choice(len_s, min(select_randomly, len_s), replace=False)
    sample_streamlines = set_number_of_points(streamlines, nb_points=nb_pts)

    if verbose:
        logger.info(f" Resampled to {nb_pts} points")
        logger.info(f" Size is {nbytes(sample_streamlines):0.3f} MB")
        logger.info(f" Duration of resampling is {time() - t:0.3f} s")
        logger.info(" QBX phase starting...")

    qbx = QuickBundlesX(thresholds, metric=AveragePointwiseEuclideanMetric())

    t1 = time()
    qbx_clusters = qbx.cluster(sample_streamlines, ordering=indices)

    if verbose:
        logger.info(" Merging phase starting ...")

    qbx_merge = QuickBundlesX(
        [thresholds[-1]], metric=AveragePointwiseEuclideanMetric()
    )

    final_level = len(thresholds)
    len_qbx_fl = len(qbx_clusters.get_clusters(final_level))
    qbx_ordering_final = rng.choice(len_qbx_fl, len_qbx_fl, replace=False)

    qbx_merged_cluster_map = qbx_merge.cluster(
        qbx_clusters.get_clusters(final_level).centroids, ordering=qbx_ordering_final
    ).get_clusters(1)

    qbx_cluster_map = qbx_clusters.get_clusters(final_level)

    merged_cluster_map = ClusterMapCentroid()
    for cluster in qbx_merged_cluster_map:
        merged_cluster = ClusterCentroid(centroid=cluster.centroid)
        for i in cluster.indices:
            merged_cluster.indices.extend(qbx_cluster_map[i].indices)
        merged_cluster_map.add_cluster(merged_cluster)

    merged_cluster_map.refdata = streamlines

    if verbose:
        logger.info(f" QuickBundlesX time for {select_randomly} random streamlines")

        logger.info(f" Duration {time() - t1:0.3f} s\n")

    return merged_cluster_map



"""
================================================
Applying image-based deformations to streamlines
================================================

This example shows how to register streamlines into a template space by
applying non-rigid deformations.

At times we will be interested in bringing a set of streamlines into some
common, reference space to compute statistics out of the registered
streamlines. For a discussion on the effects of spatial normalization
approaches on tractography the work by :footcite:t:`Greene2018` can be read.

For brevity, we will include in this example only streamlines going through
the corpus callosum connecting left to right superior frontal cortex. The
process of tracking and finding these streamlines is fully demonstrated in
the :ref:`sphx_glr_examples_built_streamline_analysis_streamline_tools.py`
example.


"""

from os.path import join as pjoin

import numpy as np

from dipy.align import affine_registration, syn_registration
from dipy.align.reslice import reslice
from dipy.core.gradients import gradient_table
from dipy.data import fetch_stanford_tracks, get_fnames
from dipy.data.fetcher import fetch_mni_template, read_mni_template
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.segment.mask import median_otsu
from dipy.tracking.streamline import transform_streamlines
from dipy.viz import has_fury, horizon, regtools

###############################################################################
# In order to get the deformation field, we will first use two b0 volumes. Both
# moving and static images are assumed to be in RAS. The first one will be the
# b0 from the Stanford HARDI dataset:

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

dwi_data, dwi_affine, dwi_img = load_nifti(hardi_fname, return_img=True)
dwi_vox_size = dwi_img.header.get_zooms()[0]
dwi_bvals, dwi_bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(dwi_bvals, bvecs=dwi_bvecs)

###############################################################################
# The second one will be the T2-contrast MNI template image. The resolution of
# the MNI template is 1x1x1 mm. However, the resolution of the Stanford HARDI
# diffusion data is 2x2x2 mm. Thus, we'll need to reslice to 2x2x2 mm
# isotropic voxel resolution to match the HARDI data.

fetch_mni_template()
img_t2_mni = read_mni_template(version="a", contrast="T2")
t2_mni_data, t2_mni_affine = img_t2_mni.get_fdata(), img_t2_mni.affine
t2_mni_voxel_size = img_t2_mni.header.get_zooms()[:3]
new_voxel_size = (2.0, 2.0, 2.0)

t2_resliced_data, t2_resliced_affine = reslice(
    t2_mni_data, t2_mni_affine, t2_mni_voxel_size, new_voxel_size
)

###############################################################################
# We remove the skull of the dwi_data. For this, we need to get the b0 volumes
# indexes.

b0_idx_stanford = np.where(gtab.b0s_mask)[0].tolist()
dwi_data_noskull, _ = median_otsu(dwi_data, vol_idx=b0_idx_stanford, numpass=6)

###############################################################################
# We filter the diffusion data from the Stanford HARDI dataset to find all the
# b0 images.

b0_data_stanford = dwi_data_noskull[..., gtab.b0s_mask]

###############################################################################
# And go on to compute the Stanford HARDI dataset mean b0 image.

mean_b0_masked_stanford = np.mean(b0_data_stanford, axis=3, dtype=dwi_data.dtype)

###############################################################################
# We will register the mean b0 to the MNI T2 image template non-rigidly to
# obtain the deformation field that will be applied to the streamlines. This is
# just one of the strategies that can be used to obtain an appropriate
# deformation field. Other strategies include computing an FA template map as
# the static image, and registering the FA map of the moving image to it. This
# may may eventually lead to results with improved accuracy, since a
# T2-contrast template image as the target for normalization does not provide
# optimal tissue contrast for maximal SyN performance.
#
# We will first perform an affine registration to roughly align the two
# volumes:

pipeline = [
    "center_of_mass",
    "translation",
    "rigid",
    "rigid_isoscaling",
    "rigid_scaling",
    "affine",
]
level_iters = [10000, 1000, 100]
sigmas = [3.0, 1.0, 0.0]
factors = [4, 2, 1]


warped_b0, warped_b0_affine = affine_registration(
    mean_b0_masked_stanford,
    t2_resliced_data,
    moving_affine=dwi_affine,
    static_affine=t2_resliced_affine,
    pipeline=pipeline,
    level_iters=level_iters,
    sigmas=sigmas,
    factors=factors,
)

###############################################################################
# We now perform the non-rigid deformation using the Symmetric Diffeomorphic
# Registration (SyN) Algorithm proposed by :footcite:t:`Avants2008` (also
# implemented in the ANTs software :footcite:p:`Avants2009`):

level_iters = [10, 10, 5]

final_warped_b0, mapping = syn_registration(
    mean_b0_masked_stanford,
    t2_resliced_data,
    moving_affine=dwi_affine,
    static_affine=t2_resliced_affine,
    prealign=warped_b0_affine,
    level_iters=level_iters,
)

###############################################################################
# We show the registration result with:

regtools.overlay_slices(
    t2_resliced_data,
    final_warped_b0,
    slice_index=None,
    slice_type=0,
    ltitle="Static",
    rtitle="Moving",
    fname="transformed_sagittal.png",
)
regtools.overlay_slices(
    t2_resliced_data,
    final_warped_b0,
    slice_index=None,
    slice_type=1,
    ltitle="Static",
    rtitle="Moving",
    fname="transformed_coronal.png",
)
regtools.overlay_slices(
    t2_resliced_data,
    final_warped_b0,
    slice_index=None,
    slice_type=2,
    ltitle="Static",
    rtitle="Moving",
    fname="transformed_axial.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Deformable registration result.
#
#
# Let's now fetch a set of streamlines from the Stanford HARDI dataset.
# Those streamlines were generated during the
# :ref:`sphx_glr_examples_built_streamline_analysis_streamline_tools.py`
# example.
#
# We read the streamlines from file in voxel space:

streamlines_files = fetch_stanford_tracks()
lr_superiorfrontal_path = pjoin(streamlines_files[1], "hardi-lr-superiorfrontal.trk")

sft = load_tractogram(lr_superiorfrontal_path, "same")

###############################################################################
# The affine registration already gives a pretty good result. We could use its
# mapping to transform the streamlines to the anatomical space (MNI T2 image).
# For that, we use `transform_streamlines` and `warped_b0_affine`. Do not
# forget that `warp_b0_affine` is the affine transformation from the mean b0
# image to the T2 template image. Thus, you typically need to apply the inverse
# transformation of the affine transformation matrix that was used to register
# the two images. This is because the transformation matrix describes how
# points in one image (mean_b0) are mapped to points in the other image
# (mni T2), and to move points from the warped_b0 space to the t2 space,
# you need to "undo" this transformation.

mni_t2_streamlines_using_affine_reg = transform_streamlines(
    sft.streamlines, np.linalg.inv(warped_b0_affine)
)

sft_in_t2_using_aff_reg = StatefulTractogram(
    mni_t2_streamlines_using_affine_reg, img_t2_mni, Space.RASMM
)

###############################################################################
# Let's visualize the streamlines in the MNI T2 space. Switch the interactive
# variable to True if you want to explore the visualization in an interactive
# window.

interactive = False

if has_fury:
    horizon(
        tractograms=[sft_in_t2_using_aff_reg],
        images=[(t2_mni_data, t2_mni_affine)],
        interactive=interactive,
        world_coords=True,
        out_png="streamlines_DSN_MNI_aff_reg.png",
    )

###############################################################################
# To get better result, we use the mapping obtain by Symmetric Diffeomorphic
# Registration (SyN). Then, we can visualize the corpus callosum bundles
# that have been deformed to adapt to the MNI template space.

mni_t2_streamlines_using_syn = mapping.transform_points_inverse(sft.streamlines)

sft_in_t2_using_syn = StatefulTractogram(
    mni_t2_streamlines_using_syn, img_t2_mni, Space.RASMM
)

if has_fury:
    horizon(
        tractograms=[sft_in_t2_using_syn],
        images=[(t2_mni_data, t2_mni_affine)],
        interactive=interactive,
        world_coords=True,
        out_png="streamlines_DSN_MNI_syn.png",
    )

###############################################################################
# Finally, we save the two registered streamlines:
#
# - `mni-lr-sft_in_t2_using_aff_reg.trk` is the streamlines registered using
#   the affine registration.
# - `sft_in_t2_using_syn` is the streamlines registered using the
#   SyN registration and prealigned with the affine registration.

save_tractogram(
    sft_in_t2_using_aff_reg,
    "mni-lr-superiorfrontal_aff_reg.trk",
    bbox_valid_check=False,
)

save_tractogram(
    sft_in_t2_using_syn, "mni-lr-superiorfrontal_syn.trk", bbox_valid_check=False
)

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



import warnings

import nibabel as nib
import numpy as np
import numpy.testing as npt

from dipy.core.sphere import HemiSphere
from dipy.data import get_fnames, get_sphere
from dipy.reconst.shm import descoteaux07_legacy_msg, sh_to_sf
from dipy.tracking import tracker
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.utils import random_seeds_from_mask


def track(method, **kwargs):
    """This tests that the number of streamlines equals the number of seeds
    when return_all=True.
    """
    fnames = get_fnames(name="disco1", include_optional=True)
    sphere = HemiSphere.from_sphere(get_sphere(name="repulsion724"))
    sh = nib.load(fnames[20]).get_fdata()
    fODFs = sh_to_sf(sh, sphere, sh_order_max=12, basis_type="tournier07", legacy=False)
    fODFs[fODFs < 0] = 0

    # seeds position and initial directions
    mask = nib.load(fnames[25]).get_fdata()
    sc = BinaryStoppingCriterion(mask)
    affine = nib.load(fnames[25]).affine
    seed_mask = np.ones(mask.shape)
    seeds = random_seeds_from_mask(
        seed_mask, affine, seeds_count=100, seed_count_per_voxel=False
    )
    directions = np.random.random(seeds.shape)
    directions = np.array([v / np.linalg.norm(v) for v in directions])

    use_sf = kwargs.get("use_sf", False)
    use_directions = kwargs.get("use_dirs", False)

    # test return_all=True
    params = {
        "max_len": 500,
        "min_len": 0,
        "step_size": 0.5,
        "voxel_size": np.ones(3),
        "max_angle": 20,
        "random_seed": 0,
        "return_all": True,
        "sf": fODFs if use_sf else None,
        "sh": sh if not use_sf else None,
        "seed_directions": directions if use_directions else None,
        "sphere": sphere,
    }
    stream_gen = method(seeds, sc, affine, **params)

    streamlines = Streamlines(stream_gen)
    if use_directions:
        npt.assert_equal(len(streamlines), len(seeds))
    else:
        npt.assert_array_less(len(streamlines), len(seeds))

    # test return_all=False
    params = {
        "max_len": 500,
        "min_len": 10,
        "step_size": 0.5,
        "voxel_size": np.ones(3),
        "max_angle": 20,
        "random_seed": 0,
        "return_all": False,
        "sf": fODFs if use_sf else None,
        "sh": sh if not use_sf else None,
        "seed_directions": directions if use_directions else None,
        "sphere": sphere,
    }

    stream_gen = method(seeds, sc, affine, **params)

    streamlines = Streamlines(stream_gen)
    npt.assert_array_less(len(streamlines), len(seeds))


def test_deterministic_tracking():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        track(tracker.deterministic_tracking, use_dirs=True)
        track(tracker.deterministic_tracking, use_sf=True, use_dirs=True)


def test_probabilistic_tracking():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        track(tracker.probabilistic_tracking, use_dirs=True)
        track(tracker.probabilistic_tracking, use_sf=True, use_dirs=True)


def test_ptt_tracking():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        track(tracker.ptt_tracking)
        track(tracker.ptt_tracking, use_sf=True, use_dirs=True)


def test_tracking_error():
    sh = np.array([(64, 61, 57)])
    seeds = [np.array([0.0, 0.0, 0.0], "float"), np.array([1.0, 2.0, 3.0], "float")]
    mask = np.ones((10, 10, 10))
    sc = BinaryStoppingCriterion(mask)

    npt.assert_raises(
        ValueError, tracker.deterministic_tracking, seeds, sc, np.eye(4), sf=sh, sh=sh
    )
    npt.assert_raises(ValueError, tracker.deterministic_tracking, seeds, sc, np.eye(4))
    npt.assert_raises(
        NotImplementedError,
        tracker.deterministic_tracking,
        seeds,
        sc,
        np.eye(4),
        peaks=sh,
    )
    npt.assert_raises(
        ValueError, tracker.deterministic_tracking, seeds, sc, np.eye(4), sf=sh
    )
    npt.assert_raises(
        ValueError,
        tracker.deterministic_tracking,
        seeds,
        sc,
        np.eye(4),
        sh=sh,
        seed_directions=1,
    )
    npt.assert_raises(
        ValueError,
        tracker.deterministic_tracking,
        seeds,
        sc,
        np.eye(4),
        sh=sh,
        seed_directions=[1],
    )
    npt.assert_raises(
        ValueError,
        tracker.deterministic_tracking,
        seeds,
        sc,
        np.eye(4),
        sf=sh,
        seed_directions=[1],
    )



import os
from os.path import join as pjoin
from tempfile import TemporaryDirectory

import numpy as np
import numpy.testing as npt
import pytest

from dipy.io.image import save_nifti
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_tractogram
from dipy.io.utils import create_nifti_header
from dipy.testing.decorators import set_random_number_generator, use_xvfb
from dipy.tracking.streamline import Streamlines
from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from dipy.viz.horizon.app import horizon
    from dipy.workflows.viz import HorizonFlow


skip_it = use_xvfb == "skip"


@pytest.mark.skipif(skip_it or not has_fury, reason="Requires FURY")
@set_random_number_generator()
def test_horizon_flow(rng):
    s1 = 10 * np.array(
        [[0, 0, 0], [1, 0, 0], [2, 0, 0], [3, 0, 0], [4, 0, 0]], dtype="f8"
    )

    s2 = 10 * np.array(
        [[0, 0, 0], [0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0]], dtype="f8"
    )

    s3 = 10 * np.array(
        [[0, 0, 0], [1, 0.2, 0], [2, 0.2, 0], [3, 0.2, 0], [4, 0.2, 0]], dtype="f8"
    )

    affine = np.array(
        [
            [1.0, 0.0, 0.0, -98.0],
            [0.0, 1.0, 0.0, -134.0],
            [0.0, 0.0, 1.0, -72.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    data = 255 * rng.random((197, 233, 189))
    vox_size = (1.0, 1.0, 1.0)

    streamlines = Streamlines()
    streamlines.append(s1)
    streamlines.append(s2)
    streamlines.append(s3)

    header = create_nifti_header(affine, data.shape, vox_size)
    sft = StatefulTractogram(streamlines, header, Space.RASMM)

    tractograms = [sft]
    images = None

    with TemporaryDirectory() as out_dir:
        horizon(
            tractograms=tractograms,
            images=images,
            cluster=True,
            cluster_thr=5,
            random_colors=False,
            length_lt=np.inf,
            length_gt=0,
            clusters_lt=np.inf,
            clusters_gt=0,
            world_coords=True,
            interactive=False,
            out_png=pjoin(out_dir, "horizon-flow.png"),
        )

        buan_colors = np.ones(streamlines.get_data().shape)

        horizon(
            tractograms=tractograms,
            buan=True,
            buan_colors=buan_colors,
            world_coords=True,
            interactive=False,
            out_png=pjoin(out_dir, "buan.png"),
        )

        data = 255 * rng.random((197, 233, 189))

        images = [(data, affine, "test/test.nii.gz")]

        horizon(
            tractograms=tractograms,
            images=images,
            cluster=True,
            cluster_thr=5,
            random_colors=False,
            length_lt=np.inf,
            length_gt=0,
            clusters_lt=np.inf,
            clusters_gt=0,
            world_coords=True,
            interactive=False,
            out_png=pjoin(out_dir, "horizon-flow-nii-images.png"),
        )

        fimg = os.path.join(out_dir, "test.nii.gz")
        ftrk = os.path.join(out_dir, "test.trk")
        fnpy = os.path.join(out_dir, "test.npy")

        save_nifti(fimg, data, affine)
        dimensions = data.shape
        nii_header = create_nifti_header(affine, dimensions, vox_size)
        sft = StatefulTractogram(streamlines, nii_header, space=Space.RASMM)
        save_tractogram(sft, ftrk, bbox_valid_check=False)

        pvalues = rng.uniform(low=0, high=1, size=(10,))
        np.save(fnpy, pvalues)

        input_files = [ftrk, fimg]

        npt.assert_equal(len(input_files), 2)

        hz_flow = HorizonFlow()

        hz_flow.run(
            input_files=input_files,
            stealth=True,
            out_dir=out_dir,
            out_stealth_png="tmp_x.png",
        )

        npt.assert_equal(os.path.exists(os.path.join(out_dir, "tmp_x.png")), True)
        npt.assert_raises(
            ValueError, hz_flow.run, input_files=input_files, bg_color=(0.2, 0.2)
        )

        hz_flow.run(
            input_files=input_files,
            stealth=True,
            bg_color=[
                0.5,
            ],
            out_dir=out_dir,
            out_stealth_png="tmp_x.png",
        )
        npt.assert_equal(os.path.exists(os.path.join(out_dir, "tmp_x.png")), True)

        input_files = [ftrk, fnpy]

        npt.assert_equal(len(input_files), 2)

        hz_flow.run(
            input_files=input_files,
            stealth=True,
            bg_color=[
                0.5,
            ],
            buan=True,
            buan_thr=0.5,
            buan_highlight=(1, 1, 0),
            out_dir=out_dir,
            out_stealth_png="tmp_x.png",
        )
        npt.assert_equal(os.path.exists(os.path.join(out_dir, "tmp_x.png")), True)

        npt.assert_raises(
            ValueError, hz_flow.run, input_files=input_files, roi_colors=(0.2, 0.2)
        )

        hz_flow.run(
            input_files=input_files,
            stealth=True,
            roi_colors=[
                0.5,
            ],
            out_dir=out_dir,
            out_stealth_png="tmp_x.png",
        )
        npt.assert_equal(os.path.exists(os.path.join(out_dir, "tmp_x.png")), True)



"""
=================================
Tractography on the DiSCo Phantom
=================================

This example compares probabilistic, deterministic and parallel
transport tractography (PTT) algorithms.

See
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py`
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_deterministic.py`
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_ptt.py`
for detailed examples of those algorithms.
"""

import os

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import binary_erosion
from scipy.stats import pearsonr

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_tractogram, save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import (
    deterministic_tracking,
    probabilistic_tracking,
    ptt_tracking,
)
from dipy.tracking.utils import connectivity_matrix, seeds_from_mask
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

print("Downloading data...")

###############################################################################
# Prepare the synthetic DiSCo data for tractography. The ground-truth
# connectome will be use to evaluate tractography performances.
fnames = get_fnames(name="disco1")
disco1_fnames = [os.path.basename(f) for f in fnames]

GT_connectome_fname = fnames[
    disco1_fnames.index("DiSCo1_Connectivity_Matrix_Cross-Sectional_Area.txt")
]
GT_connectome = np.loadtxt(GT_connectome_fname)

# select the non-zero connections of the upper triangular part of the connectome
connectome_mask = np.tril(np.ones(GT_connectome.shape), -1) > 0

# load the
labels_fname = fnames[disco1_fnames.index("highRes_DiSCo1_ROIs.nii.gz")]
labels, affine, labels_img = load_nifti(labels_fname, return_img=True)
labels = labels.astype(int)

print("Loading data...")
tracks_fname = fnames[disco1_fnames.index("DiSCo1_Strands_Trajectories.tck")]
GT_streams = load_tractogram(tracks_fname, reference=labels_img).streamlines
if has_fury:
    scene = window.Scene()
    scene.add(actor.line(GT_streams, colors=colormap.line_colors(GT_streams)))
    window.record(scene=scene, out_path="tractogram_ground_truth.png", size=(800, 800))
    if interactive:
        window.show(scene)

plt.imshow(GT_connectome, origin="lower", cmap="viridis", interpolation="nearest")
plt.axis("off")
plt.savefig("connectome_ground_truth.png")
plt.close()

###############################################################################
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# DiSCo ground-truth trajectories (left) and connectivity matrix (right).
#
#

# Tracking mask, seed positions and initial directions
mask_fname = fnames[disco1_fnames.index("highRes_DiSCo1_mask.nii.gz")]
mask = load_nifti_data(mask_fname)
sc = BinaryStoppingCriterion(mask)

voxel_size = np.ones(3)
seed_fname = fnames[disco1_fnames.index("highRes_DiSCo1_ROIs-mask.nii.gz")]
seed_mask = load_nifti_data(seed_fname)
seed_mask = binary_erosion(seed_mask * mask, iterations=1)
seeds = seeds_from_mask(seed_mask, affine, density=2)

plt.imshow(seed_mask[:, :, 17], origin="lower", cmap="gray", interpolation="nearest")
plt.axis("off")
plt.title("Seeding Mask")
plt.savefig("seeding_mask.png")
plt.close()
plt.imshow(mask[:, :, 17], origin="lower", cmap="gray", interpolation="nearest")
plt.axis("off")
plt.title("Tracking Mask")
plt.savefig("tracking_mask.png")
plt.close()

###############################################################################
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# DiSCo seeding (left) and tracking (right) masks.
#
#

# Compute ODFs
data_fname = fnames[disco1_fnames.index("highRes_DiSCo1_DWI_RicianNoise-snr10.nii.gz")]
data = load_nifti_data(data_fname)

bvecs = fnames[disco1_fnames.index("DiSCo_gradients_dipy.bvecs")]
bvals = fnames[disco1_fnames.index("DiSCo_gradients.bvals")]
gtab = gradient_table(bvals=bvals, bvecs=bvecs)

response, _ = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=8)
csd_fit = csd_model.fit(data, mask=mask)
ODFs = csd_fit.odf(default_sphere)

if has_fury:
    scene = window.Scene()
    ODF_spheres = actor.odf_slicer(
        ODFs[:, :, 17:18, :],
        sphere=default_sphere,
        scale=2,
        norm=False,
        colormap="plasma",
    )
    scene.add(ODF_spheres)
    window.record(scene=scene, out_path="GT_odfs.png", size=(600, 600))
    if interactive:
        window.show(scene)

###############################################################################
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# DiSCo ground-truth ODFs.
#
#

# Perform deterministic tractography using 1 thread (cpu)
print("Running Deterministic Tractography...")
streamline_generator = deterministic_tracking(
    seeds, sc, affine, sf=ODFs, nbr_threads=1, random_seed=42, sphere=default_sphere
)

det_streams = Streamlines(streamline_generator)
sft = StatefulTractogram(det_streams, labels_img, Space.RASMM)
save_trk(sft, "tractogram_disco_deterministic.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(det_streams, colors=colormap.line_colors(det_streams)))
    window.record(
        scene=scene, out_path="tractogram_disco_deterministic.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

# Compare the estimated connectivity with the ground-truth connectivity
connectome = connectivity_matrix(det_streams, affine, labels)[1:, 1:]
r, _ = pearsonr(
    GT_connectome[connectome_mask].flatten(), connectome[connectome_mask].flatten()
)
print("DiSCo ground-truth correlation (deterministic tractography): ", r)

plt.imshow(connectome, origin="lower", cmap="viridis", interpolation="nearest")
plt.axis("off")
plt.savefig("connectome_deterministic.png")
plt.close()

###############################################################################
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# DiSCo Deterministic tractogram and corresponding connectome.
#
#
# Perform probabilistic tractography using 4 threads (cpus)

print("Running Probabilistic Tractography...")
streamline_generator = probabilistic_tracking(
    seeds, sc, affine, sf=ODFs, nbr_threads=4, random_seed=42, sphere=default_sphere
)
prob_streams = Streamlines(streamline_generator)
sft = StatefulTractogram(prob_streams, labels_img, Space.RASMM)
save_trk(sft, "tractogram_disco_probabilistic.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(prob_streams, colors=colormap.line_colors(prob_streams)))
    window.record(
        scene=scene, out_path="tractogram_disco_probabilistic.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

# Compare the estimated connectivity with the ground-truth connectivity
connectome = connectivity_matrix(prob_streams, affine, labels)[1:, 1:]
r, _ = pearsonr(
    GT_connectome[connectome_mask].flatten(), connectome[connectome_mask].flatten()
)
print("DiSCo ground-truth correlation (probabilistic tractography): ", r)

plt.imshow(connectome, origin="lower", cmap="viridis", interpolation="nearest")
plt.axis("off")
plt.savefig("connectome_probabilistic.png")
plt.close()

###############################################################################
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# DiSCo Probabilistic tractogram and corresponding connectome.
#
#

# Perform parallel transport tractography tractography using all threads (cpus)
print("Running Parallel Transport Tractography...")
streamline_generator = ptt_tracking(
    seeds,
    sc,
    affine,
    sf=ODFs,
    nbr_threads=0,
    random_seed=42,
    sphere=default_sphere,
    max_angle=20,
)
ptt_streams = Streamlines(streamline_generator)
sft = StatefulTractogram(ptt_streams, labels_img, Space.RASMM)
save_trk(sft, "tractogram_disco_ptt.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(ptt_streams, colors=colormap.line_colors(ptt_streams)))
    window.record(scene=scene, out_path="tractogram_disco_ptt.png", size=(800, 800))
    if interactive:
        window.show(scene)

# Compare the estimated connectivity with the ground-truth connectivity
connectome = connectivity_matrix(ptt_streams, affine, labels)[1:, 1:]
r, _ = pearsonr(
    GT_connectome[connectome_mask].flatten(), connectome[connectome_mask].flatten()
)
print("DiSCo ground-truth correlation (PTT tractography): ", r)
plt.imshow(connectome, origin="lower", cmap="viridis", interpolation="nearest")
plt.axis("off")
plt.savefig("connectome_ptt.png")
plt.close()

###############################################################################
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# DiSCo PTT tractogram and corresponding connectome.
#
#



import numpy as np
from numpy.testing import (
    assert_,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_equal,
    assert_raises,
)

from dipy.align.bundlemin import (
    _bundle_minimum_distance,
    _bundle_minimum_distance_matrix,
    distance_matrix_mdf,
)
from dipy.align.streamlinear import (
    BundleMinDistanceMatrixMetric,
    BundleMinDistanceMetric,
    BundleSumDistanceMatrixMetric,
    StreamlineDistanceMetric,
    StreamlineLinearRegistration,
    compose_matrix44,
    decompose_matrix44,
    get_unique_pairs,
    groupwise_slr,
)
from dipy.core.geometry import compose_matrix
from dipy.data import get_fnames, read_five_af_bundles, two_cingulum_bundles
from dipy.io.streamline import load_tractogram
from dipy.testing.decorators import set_random_number_generator
from dipy.tracking.streamline import (
    Streamlines,
    center_streamlines,
    relist_streamlines,
    set_number_of_points,
    transform_streamlines,
    unlist_streamlines,
)


def simulated_bundle(no_streamlines=10, waves=False, no_pts=12):
    t = np.linspace(-10, 10, 200)
    # parallel waves or parallel lines
    bundle = []
    for i in np.linspace(-5, 5, no_streamlines):
        if waves:
            pts = np.vstack((np.cos(t), t, i * np.ones(t.shape))).T
        else:
            pts = np.vstack((np.zeros(t.shape), t, i * np.ones(t.shape))).T
        pts = set_number_of_points(pts, nb_points=no_pts)
        bundle.append(pts)

    return bundle


def fornix_streamlines(no_pts=12):
    fname = get_fnames(name="fornix")

    fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

    fornix_streamlines = Streamlines(fornix)
    streamlines = set_number_of_points(fornix_streamlines, nb_points=no_pts)
    return streamlines


def evaluate_convergence(bundle, new_bundle2):
    pts_static = np.concatenate(bundle, axis=0)
    pts_moved = np.concatenate(new_bundle2, axis=0)
    assert_array_almost_equal(pts_static, pts_moved, 3)


def test_rigid_parallel_lines():
    bundle_initial = simulated_bundle()
    bundle, shift = center_streamlines(bundle_initial)
    mat = compose_matrix44([20, 0, 10, 0, 40, 0])

    bundle2 = transform_streamlines(bundle, mat)

    bundle_sum_distance = BundleSumDistanceMatrixMetric()
    options = {"maxcor": 100, "ftol": 1e-9, "gtol": 1e-16, "eps": 1e-3}
    srr = StreamlineLinearRegistration(
        metric=bundle_sum_distance,
        x0=np.zeros(6),
        method="L-BFGS-B",
        bounds=None,
        options=options,
    )

    new_bundle2 = srr.optimize(bundle, bundle2).transform(bundle2)
    evaluate_convergence(bundle, new_bundle2)


def test_rigid_real_bundles():
    bundle_initial = fornix_streamlines()[:20]
    bundle, shift = center_streamlines(bundle_initial)

    mat = compose_matrix44([0, 0, 20, 45.0, 0, 0])

    bundle2 = transform_streamlines(bundle, mat)

    bundle_sum_distance = BundleSumDistanceMatrixMetric()
    srr = StreamlineLinearRegistration(
        metric=bundle_sum_distance, x0=np.zeros(6), method="Powell"
    )
    new_bundle2 = srr.optimize(bundle, bundle2).transform(bundle2)

    evaluate_convergence(bundle, new_bundle2)

    bundle_min_distance = BundleMinDistanceMatrixMetric()
    srr = StreamlineLinearRegistration(
        metric=bundle_min_distance, x0=np.zeros(6), method="Powell"
    )
    new_bundle2 = srr.optimize(bundle, bundle2).transform(bundle2)

    evaluate_convergence(bundle, new_bundle2)

    assert_raises(ValueError, StreamlineLinearRegistration, method="Whatever")


def test_rigid_partial_real_bundles():
    static = fornix_streamlines()[:20]
    moving = fornix_streamlines()[20:40]
    static_center, shift = center_streamlines(static)
    moving_center, shift2 = center_streamlines(moving)

    print(shift2)
    mat = compose_matrix(
        translate=np.array([0, 0, 0.0]), angles=np.deg2rad([40, 0, 0.0])
    )
    moved = transform_streamlines(moving_center, mat)

    srr = StreamlineLinearRegistration()

    srm = srr.optimize(static_center, moved)
    print(srm.fopt)
    print(srm.iterations)
    print(srm.funcs)

    moving_back = srm.transform(moved)
    print(srm.matrix)

    static_center = set_number_of_points(static_center, nb_points=100)
    moving_center = set_number_of_points(moving_back, nb_points=100)

    vol = np.zeros((100, 100, 100))
    spts = np.concatenate(static_center, axis=0)
    spts = np.round(spts).astype(int) + np.array([50, 50, 50])

    mpts = np.concatenate(moving_center, axis=0)
    mpts = np.round(mpts).astype(int) + np.array([50, 50, 50])

    for index in spts:
        i, j, k = index
        vol[i, j, k] = 1

    vol2 = np.zeros((100, 100, 100))
    for index in mpts:
        i, j, k = index
        vol2[i, j, k] = 1

    overlap = np.sum(np.logical_and(vol, vol2)) / float(np.sum(vol2))

    assert_equal(overlap * 100 > 40, True)


def test_stream_rigid():
    static = fornix_streamlines()[:20]
    moving = fornix_streamlines()[20:40]
    center_streamlines(static)

    mat = compose_matrix44([0, 0, 0, 0, 40, 0])
    moving = transform_streamlines(moving, mat)

    srr = StreamlineLinearRegistration()
    sr_params = srr.optimize(static, moving)
    moved = transform_streamlines(moving, sr_params.matrix)

    srr = StreamlineLinearRegistration(verbose=True)
    srm = srr.optimize(static, moving)
    moved2 = transform_streamlines(moving, srm.matrix)
    moved3 = srm.transform(moving)

    assert_array_almost_equal(moved[0], moved2[0], decimal=3)
    assert_array_almost_equal(moved2[0], moved3[0], decimal=3)


def test_min_vs_min_fast_precision():
    static = fornix_streamlines()[:20]
    moving = fornix_streamlines()[:20]

    static = [s.astype("f8") for s in static]
    moving = [m.astype("f8") for m in moving]

    bmd = BundleMinDistanceMatrixMetric()
    bmd.setup(static, moving)

    bmdf = BundleMinDistanceMetric()
    bmdf.setup(static, moving)

    x_test = [0.01, 0, 0, 0, 0, 0]

    print(bmd.distance(x_test))
    print(bmdf.distance(x_test))
    assert_equal(bmd.distance(x_test), bmdf.distance(x_test))


@set_random_number_generator()
def test_same_number_of_points(rng):
    A = [rng.random((10, 3)), rng.random((20, 3))]
    B = [rng.random((21, 3)), rng.random((30, 3))]
    C = [rng.random((10, 3)), rng.random((10, 3))]
    D = [rng.random((20, 3)), rng.random((20, 3))]

    slr = StreamlineLinearRegistration()
    assert_raises(ValueError, slr.optimize, A, B)
    assert_raises(ValueError, slr.optimize, C, D)
    assert_raises(ValueError, slr.optimize, C, B)


def test_efficient_bmd():
    a = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])

    streamlines = [a, a + 2, a + 4]

    points, offsets = unlist_streamlines(streamlines)
    points = points.astype(np.double)
    points2 = points.copy()

    D = np.zeros((len(offsets), len(offsets)), dtype="f8")

    _bundle_minimum_distance_matrix(
        points, points2, len(offsets), len(offsets), a.shape[0], D
    )

    assert_equal(np.sum(np.diag(D)), 0)

    points2 += 2

    _bundle_minimum_distance_matrix(
        points, points2, len(offsets), len(offsets), a.shape[0], D
    )

    streamlines2 = relist_streamlines(points2, offsets)
    D2 = distance_matrix_mdf(streamlines, streamlines2)

    assert_array_almost_equal(D, D2)

    cols = D2.shape[1]
    rows = D2.shape[0]

    dist = (
        0.25
        * (
            np.sum(np.min(D2, axis=0)) / float(cols)
            + np.sum(np.min(D2, axis=1)) / float(rows)
        )
        ** 2
    )

    dist2 = _bundle_minimum_distance(
        points, points2, len(offsets), len(offsets), a.shape[0]
    )
    assert_almost_equal(dist, dist2)


@set_random_number_generator()
def test_openmp_locks(rng):
    static = []
    moving = []
    pts = 20

    for _ in range(1000):
        s = rng.random((pts, 3))
        static.append(s)
        moving.append(s + 2)

    moving = moving[2:]

    points, offsets = unlist_streamlines(static)
    points2, offsets2 = unlist_streamlines(moving)

    D = np.zeros((len(offsets), len(offsets2)), dtype="f8")

    _bundle_minimum_distance_matrix(
        points, points2, len(offsets), len(offsets2), pts, D
    )

    dist1 = (
        0.25
        * (
            np.sum(np.min(D, axis=0)) / float(D.shape[1])
            + np.sum(np.min(D, axis=1)) / float(D.shape[0])
        )
        ** 2
    )

    dist2 = _bundle_minimum_distance(points, points2, len(offsets), len(offsets2), pts)

    assert_almost_equal(dist1, dist2, 6)


def test_from_to_rigid():
    t = np.array([10, 2, 3, 0.1, 20.0, 30.0])
    mat = compose_matrix44(t)
    vec = decompose_matrix44(mat, size=6)

    assert_array_almost_equal(t, vec)

    t = np.array([0, 0, 0, 180, 0.0, 0.0])

    mat = np.eye(4)
    mat[0, 0] = -1

    vec = decompose_matrix44(mat, size=6)

    assert_array_almost_equal(-t, vec)


def test_matrix44():
    assert_raises(ValueError, compose_matrix44, np.ones(5))
    assert_raises(ValueError, compose_matrix44, np.ones(13))
    assert_raises(ValueError, compose_matrix44, np.ones(16))


def test_abstract_metric_class():
    class DummyStreamlineMetric(StreamlineDistanceMetric):
        def test(self):
            pass

    assert_raises(TypeError, DummyStreamlineMetric)


def test_evolution_of_previous_iterations():
    static = fornix_streamlines()[:20]
    moving = fornix_streamlines()[:20]

    moving = [m + np.array([10.0, 0.0, 0.0]) for m in moving]

    slr = StreamlineLinearRegistration(evolution=True)

    slm = slr.optimize(static, moving)

    assert_equal(len(slm.matrix_history), slm.iterations)


def test_similarity_real_bundles():
    bundle_initial = fornix_streamlines()
    bundle_initial, shift = center_streamlines(bundle_initial)
    bundle = bundle_initial[:20]
    xgold = [0, 0, 10, 0, 0, 0, 1.5]
    mat = compose_matrix44(xgold)
    bundle2 = transform_streamlines(bundle_initial[:20], mat)

    metric = BundleMinDistanceMatrixMetric()
    x0 = np.array([0, 0, 0, 0, 0, 0, 1], "f8")

    slr = StreamlineLinearRegistration(
        metric=metric, x0=x0, method="Powell", bounds=None, verbose=False
    )

    slm = slr.optimize(bundle, bundle2)
    new_bundle2 = slm.transform(bundle2)
    evaluate_convergence(bundle, new_bundle2)


def test_affine_real_bundles():
    bundle_initial = fornix_streamlines()
    bundle_initial, shift = center_streamlines(bundle_initial)
    bundle = bundle_initial[:20]
    xgold = [0, 4, 2, 0, 10, 10, 1.2, 1.1, 1.0, 0.0, 0.2, 0.0]
    mat = compose_matrix44(xgold)
    bundle2 = transform_streamlines(bundle_initial[:20], mat)

    x0 = np.array([0, 0, 0, 0, 0, 0, 1.0, 1.0, 1.0, 0, 0, 0])

    x = 25

    bounds = [
        (-x, x),
        (-x, x),
        (-x, x),
        (-x, x),
        (-x, x),
        (-x, x),
        (0.1, 1.5),
        (0.1, 1.5),
        (0.1, 1.5),
        (-1, 1),
        (-1, 1),
        (-1, 1),
    ]

    options = {"maxcor": 10, "ftol": 1e-7, "gtol": 1e-5, "eps": 1e-8}

    metric = BundleMinDistanceMatrixMetric()

    slr = StreamlineLinearRegistration(
        metric=metric,
        x0=x0,
        method="L-BFGS-B",
        bounds=bounds,
        verbose=True,
        options=options,
    )
    slm = slr.optimize(bundle, bundle2)

    new_bundle2 = slm.transform(bundle2)

    slr2 = StreamlineLinearRegistration(
        metric=metric, x0=x0, method="Powell", bounds=None, verbose=True, options=None
    )

    slm2 = slr2.optimize(bundle, new_bundle2)

    new_bundle2 = slm2.transform(new_bundle2)

    evaluate_convergence(bundle, new_bundle2)


def test_vectorize_streamlines():
    cingulum_bundles = two_cingulum_bundles()

    cb_subj1 = cingulum_bundles[0]
    cb_subj1 = set_number_of_points(cb_subj1, nb_points=10)
    cb_subj1_pts_no = np.array([s.shape[0] for s in cb_subj1])

    assert_equal(np.all(cb_subj1_pts_no == 10), True)


@set_random_number_generator()
def test_x0_input(rng):
    for x0 in [6, 7, 12, "Rigid", "rigid", "similarity", "Affine"]:
        StreamlineLinearRegistration(x0=x0)

    for x0 in [rng.random(6), rng.random(7), rng.random(12)]:
        StreamlineLinearRegistration(x0=x0)

    for x0 in [8, 20, "Whatever", rng.random(20), rng.random((20, 3))]:
        assert_raises(ValueError, StreamlineLinearRegistration, x0=x0)

    x0 = rng.random((4, 3))
    assert_raises(ValueError, StreamlineLinearRegistration, x0=x0)

    x0_6 = np.zeros(6)
    x0_7 = np.array([0, 0, 0, 0, 0, 0, 1.0])
    x0_12 = np.array([0, 0, 0, 0, 0, 0, 1.0, 1.0, 1.0, 0, 0, 0])

    x0_s = [x0_6, x0_7, x0_12, x0_6, x0_7, x0_12]

    for i, x0 in enumerate([6, 7, 12, "Rigid", "similarity", "Affine"]):
        slr = StreamlineLinearRegistration(x0=x0)
        assert_equal(slr.x0, x0_s[i])


@set_random_number_generator()
def test_compose_decompose_matrix44(rng):
    for _ in range(20):
        x0 = rng.random(12)
        mat = compose_matrix44(x0[:6])
        assert_array_almost_equal(x0[:6], decompose_matrix44(mat, size=6))
        mat = compose_matrix44(x0[:7])
        assert_array_almost_equal(x0[:7], decompose_matrix44(mat, size=7))
        mat = compose_matrix44(x0[:12])
        assert_array_almost_equal(x0[:12], decompose_matrix44(mat, size=12))

    assert_raises(ValueError, decompose_matrix44, mat, size=20)


def test_cascade_of_optimizations_and_threading():
    cingulum_bundles = two_cingulum_bundles()

    cb1 = cingulum_bundles[0]
    cb1 = set_number_of_points(cb1, nb_points=20)

    test_x0 = np.array([10, 4, 3, 0, 20, 10, 1.5, 1.5, 1.5, 0.0, 0.2, 0])

    cb2 = transform_streamlines(cingulum_bundles[0], compose_matrix44(test_x0))
    cb2 = set_number_of_points(cb2, nb_points=20)

    print("first rigid")
    slr = StreamlineLinearRegistration(x0=6, num_threads=1)
    slm = slr.optimize(cb1, cb2)

    print("then similarity")
    slr2 = StreamlineLinearRegistration(x0=7, num_threads=2)
    slm2 = slr2.optimize(cb1, cb2, mat=slm.matrix)

    print("then affine")
    slr3 = StreamlineLinearRegistration(
        x0=12, options={"maxiter": 50}, num_threads=None
    )
    slm3 = slr3.optimize(cb1, cb2, mat=slm2.matrix)

    assert_(slm2.fopt < slm.fopt)
    assert_(slm3.fopt < slm2.fopt)


@set_random_number_generator()
def test_wrong_num_threads(rng):
    A = [rng.random((10, 3)), rng.random((10, 3))]
    B = [rng.random((10, 3)), rng.random((10, 3))]

    slr = StreamlineLinearRegistration(num_threads=0)
    assert_raises(ValueError, slr.optimize, A, B)


def test_get_unique_pairs():
    # Regular case
    pairs, exclude = get_unique_pairs(6)
    assert_equal(len(np.unique(pairs)), 6)
    assert_equal(exclude, None)

    # Odd case
    pairs, exclude = get_unique_pairs(5)
    assert_equal(len(np.unique(pairs)), 4)
    assert_equal(isinstance(exclude, (int, np.int64, np.int32)), True)

    # Iterative case
    new_pairs, new_exclude = get_unique_pairs(5, pairs=pairs)
    assert_equal(len(np.unique(pairs)), 4)
    assert_equal(exclude != new_exclude, True)

    # Check errors
    assert_raises(TypeError, get_unique_pairs, 2.7)
    assert_raises(ValueError, get_unique_pairs, 1)


def test_groupwise_slr():
    bundles = read_five_af_bundles()

    # Test regular use case with convergence
    new_bundles, T, d = groupwise_slr(bundles, verbose=True)

    assert_equal(len(new_bundles), len(bundles))
    assert_equal(type(new_bundles), list)
    assert_equal(len(T), len(bundles))
    assert_equal(type(T), list)

    # Test regular use case without convergence (few iterations)
    new_bundles, T, d = groupwise_slr(bundles, max_iter=3, tol=-10, verbose=True)



"""
===========================
Calculate SHORE scalar maps
===========================

We show how to calculate two SHORE-based scalar maps: return to origin
probability (RTOP) :footcite:p:`Descoteaux2011` and mean square displacement
(MSD) :footcite:p:`Wu2007`, :footcite:p:`Wu2008` on your data. SHORE can be used
with any multiple b-value dataset like multi-shell or DSI.

First import the necessary modules:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.shore import ShoreModel

###############################################################################
# Download and get the data filenames for this tutorial.

fraw, fbval, fbvec = get_fnames(name="taiwan_ntu_dsi")

###############################################################################
# img contains a nibabel Nifti1Image object (data) and gtab contains a
# GradientTable object (gradient information e.g. b-values). For example, to
# read the b-values it is possible to write print(gtab.bvals).
#
# Load the raw diffusion data and the affine.

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
bvecs[1:] = bvecs[1:] / np.sqrt(np.sum(bvecs[1:] * bvecs[1:], axis=1))[:, None]
gtab = gradient_table(bvals, bvecs=bvecs)
print(f"data.shape {data.shape}")

###############################################################################
# Instantiate the Model.

asm = ShoreModel(gtab)

###############################################################################
# Let's just use only one slice only from the data.

dataslice = data[30:70, 20:80, data.shape[2] // 2]

###############################################################################
# Fit the signal with the model and calculate the SHORE coefficients.

asmfit = asm.fit(dataslice)

###############################################################################
# Calculate the analytical RTOP on the signal
# that corresponds to the integral of the signal.

print("Calculating... rtop_signal")
rtop_signal = asmfit.rtop_signal()

###############################################################################
# Now we calculate the analytical RTOP on the propagator,
# that corresponds to its central value.

print("Calculating... rtop_pdf")
rtop_pdf = asmfit.rtop_pdf()

###############################################################################
# In theory, these two measures must be equal,
# to show that we calculate the mean square error on this two measures.

mse = np.sum((rtop_signal - rtop_pdf) ** 2) / rtop_signal.size
print(f"MSE = {mse:f}")

###############################################################################
# Let's calculate the analytical mean square displacement on the propagator.

print("Calculating... msd")
msd = asmfit.msd()

###############################################################################
# Show the maps and save them to a file.

fig = plt.figure(figsize=(6, 6))
ax1 = fig.add_subplot(2, 2, 1, title="rtop_signal")
ax1.set_axis_off()
ind = ax1.imshow(rtop_signal.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
ax2 = fig.add_subplot(2, 2, 2, title="rtop_pdf")
ax2.set_axis_off()
ind = ax2.imshow(rtop_pdf.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
ax3 = fig.add_subplot(2, 2, 3, title="msd")
ax3.set_axis_off()
ind = ax3.imshow(msd.T, interpolation="nearest", origin="lower", vmin=0)
plt.colorbar(ind)
plt.savefig("SHORE_maps.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RTOP and MSD calculated using the SHORE model.
#
#
# References
# ----------
#
# .. footbibliography::
#



from itertools import repeat
import multiprocessing as mp
from os import path
import tempfile
import warnings

import numpy as np
import scipy.optimize as opt

from dipy.core.interpolation import trilinear_interpolate4d
from dipy.core.ndindex import ndindex
from dipy.core.sphere import Sphere
from dipy.data import default_sphere
from dipy.reconst.dirspeed import peak_directions
from dipy.reconst.eudx_direction_getter import EuDXDirectionGetter
from dipy.reconst.odf import gfa
from dipy.reconst.recspeed import (
    local_maxima,
    remove_similar_vertices,
    search_descending,
)
from dipy.reconst.shm import sh_to_sf_matrix
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.deprecator import deprecated_params
from dipy.utils.multiproc import determine_num_processes


@warning_for_keywords()
def peak_directions_nl(
    sphere_eval,
    *,
    relative_peak_threshold=0.25,
    min_separation_angle=25,
    sphere=default_sphere,
    xtol=1e-7,
):
    """Non Linear Direction Finder.

    Parameters
    ----------
    sphere_eval : callable
        A function which can be evaluated on a sphere.
    relative_peak_threshold : float
        Only return peaks greater than ``relative_peak_threshold * m`` where m
        is the largest peak.
    min_separation_angle : float in [0, 90]
        The minimum distance between directions. If two peaks are too close
        only the larger of the two is returned.
    sphere : Sphere
        A discrete Sphere. The points on the sphere will be used for initial
        estimate of maximums.
    xtol : float
        Relative tolerance for optimization.

    Returns
    -------
    directions : array (N, 3)
        Points on the sphere corresponding to N local maxima on the sphere.
    values : array (N,)
        Value of sphere_eval at each point on directions.

    """
    # Find discrete peaks for use as seeds in non-linear search
    discrete_values = sphere_eval(sphere)
    values, indices = local_maxima(discrete_values, sphere.edges)

    seeds = np.column_stack([sphere.theta[indices], sphere.phi[indices]])

    # Helper function
    def _helper(x):
        sphere = Sphere(theta=x[0], phi=x[1])
        return -sphere_eval(sphere)

    # Non-linear search
    num_seeds = len(seeds)
    theta = np.empty(num_seeds)
    phi = np.empty(num_seeds)
    for i in range(num_seeds):
        peak = opt.fmin(_helper, seeds[i], xtol=xtol, disp=False)
        theta[i], phi[i] = peak

    # Evaluate on new-found peaks
    small_sphere = Sphere(theta=theta, phi=phi)
    values = sphere_eval(small_sphere)

    # Sort in descending order
    order = values.argsort()[::-1]
    values = values[order]
    directions = small_sphere.vertices[order]

    # Remove directions that are too small
    n = search_descending(values, relative_peak_threshold)
    directions = directions[:n]

    # Remove peaks too close to each-other
    directions, idx = remove_similar_vertices(
        directions, min_separation_angle, return_index=True
    )
    values = values[idx]
    return directions, values


def _pam_from_attrs(
    klass, sphere, peak_indices, peak_values, peak_dirs, gfa, qa, shm_coeff, B, odf
):
    """
    Construct PeaksAndMetrics object (or subclass) from its attributes.

    This is also useful for pickling/unpickling of these objects (see also
    :func:`__reduce__` below).

    Parameters
    ----------
    klass : class
        The class of object to be created.
    sphere : `Sphere` class instance.
        Sphere for discretization.
    peak_indices : ndarray
        Indices (in sphere vertices) of the peaks in each voxel.
    peak_values : ndarray
        The value of the peaks.
    peak_dirs : ndarray
        The direction of each peak.
    gfa : ndarray
        The Generalized Fractional Anisotropy in each voxel.
    qa : ndarray
        Quantitative Anisotropy in each voxel.
    shm_coeff : ndarray
        The coefficients of the spherical harmonic basis for the ODF in
        each voxel.
    B : ndarray
        The spherical harmonic matrix, for multiplication with the
        coefficients.
    odf : ndarray
        The orientation distribution function on the sphere in each voxel.

    Returns
    -------
    pam : Instance of the class `klass`.
    """
    this_pam = klass()
    this_pam.sphere = sphere
    this_pam.peak_dirs = peak_dirs
    this_pam.peak_values = peak_values
    this_pam.peak_indices = peak_indices
    this_pam.gfa = gfa
    this_pam.qa = qa
    this_pam.shm_coeff = shm_coeff
    this_pam.B = B
    this_pam.odf = odf
    return this_pam


class PeaksAndMetrics(EuDXDirectionGetter):
    def __reduce__(self):
        return _pam_from_attrs, (
            self.__class__,
            self.sphere,
            self.peak_indices,
            self.peak_values,
            self.peak_dirs,
            self.gfa,
            self.qa,
            self.shm_coeff,
            self.B,
            self.odf,
        )


def _peaks_from_model_parallel(
    model,
    data,
    sphere,
    relative_peak_threshold,
    min_separation_angle,
    mask,
    return_odf,
    return_sh,
    gfa_thr,
    normalize_peaks,
    sh_order,
    sh_basis_type,
    legacy,
    npeaks,
    B,
    invB,
    num_processes,
):
    shape = list(data.shape)
    data = np.reshape(data, (-1, shape[-1]))
    n = data.shape[0]
    nbr_chunks = num_processes**2
    chunk_size = int(np.ceil(n / nbr_chunks))
    indices = list(
        zip(
            np.arange(0, n, chunk_size),
            np.arange(0, n, chunk_size) + chunk_size,
        )
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        data_file_name = path.join(tmpdir, "data.npy")
        np.save(data_file_name, data)
        if mask is not None:
            mask = mask.flatten()
            mask_file_name = path.join(tmpdir, "mask.npy")
            np.save(mask_file_name, mask)
        else:
            mask_file_name = None

        mp.set_start_method("spawn", force=True)
        pool = mp.Pool(num_processes)

        pam_res = pool.map(
            _peaks_from_model_parallel_sub,
            zip(
                repeat((data_file_name, mask_file_name)),
                indices,
                repeat(model),
                repeat(sphere),
                repeat(relative_peak_threshold),
                repeat(min_separation_angle),
                repeat(return_odf),
                repeat(return_sh),
                repeat(gfa_thr),
                repeat(normalize_peaks),
                repeat(sh_order),
                repeat(sh_basis_type),
                repeat(legacy),
                repeat(npeaks),
                repeat(B),
                repeat(invB),
            ),
        )
        pool.close()

        pam = PeaksAndMetrics()
        pam.sphere = sphere
        # use memmap to reduce the memory usage
        pam.gfa = np.memmap(
            path.join(tmpdir, "gfa.npy"),
            dtype=pam_res[0].gfa.dtype,
            mode="w+",
            shape=(data.shape[0]),
        )

        pam.peak_dirs = np.memmap(
            path.join(tmpdir, "peak_dirs.npy"),
            dtype=pam_res[0].peak_dirs.dtype,
            mode="w+",
            shape=(data.shape[0], npeaks, 3),
        )
        pam.peak_values = np.memmap(
            path.join(tmpdir, "peak_values.npy"),
            dtype=pam_res[0].peak_values.dtype,
            mode="w+",
            shape=(data.shape[0], npeaks),
        )
        pam.peak_indices = np.memmap(
            path.join(tmpdir, "peak_indices.npy"),
            dtype=pam_res[0].peak_indices.dtype,
            mode="w+",
            shape=(data.shape[0], npeaks),
        )
        pam.qa = np.memmap(
            path.join(tmpdir, "qa.npy"),
            dtype=pam_res[0].qa.dtype,
            mode="w+",
            shape=(data.shape[0], npeaks),
        )
        if return_sh:
            nbr_shm_coeff = (sh_order + 2) * (sh_order + 1) // 2
            pam.shm_coeff = np.memmap(
                path.join(tmpdir, "shm.npy"),
                dtype=pam_res[0].shm_coeff.dtype,
                mode="w+",
                shape=(data.shape[0], nbr_shm_coeff),
            )
            pam.B = pam_res[0].B
        else:
            pam.shm_coeff = None
            pam.invB = None
        if return_odf:
            pam.odf = np.memmap(
                path.join(tmpdir, "odf.npy"),
                dtype=pam_res[0].odf.dtype,
                mode="w+",
                shape=(data.shape[0], len(sphere.vertices)),
            )
        else:
            pam.odf = None

        # copy subprocesses pam to a single pam (memmaps)
        for i, (start_pos, end_pos) in enumerate(indices):
            pam.gfa[start_pos:end_pos] = pam_res[i].gfa
            pam.peak_dirs[start_pos:end_pos] = pam_res[i].peak_dirs
            pam.peak_values[start_pos:end_pos] = pam_res[i].peak_values
            pam.peak_indices[start_pos:end_pos] = pam_res[i].peak_indices
            pam.qa[start_pos:end_pos] = pam_res[i].qa
            if return_sh:
                pam.shm_coeff[start_pos:end_pos] = pam_res[i].shm_coeff
            if return_odf:
                pam.odf[start_pos:end_pos] = pam_res[i].odf

        # load memmaps to arrays and reshape the metric
        shape[-1] = -1
        pam.gfa = np.reshape(np.array(pam.gfa), shape[:-1])
        pam.peak_dirs = np.reshape(np.array(pam.peak_dirs), shape + [3])
        pam.peak_values = np.reshape(np.array(pam.peak_values), shape)
        pam.peak_indices = np.reshape(np.array(pam.peak_indices), shape)
        pam.qa = np.reshape(np.array(pam.qa), shape)
        if return_sh:
            pam.shm_coeff = np.reshape(np.array(pam.shm_coeff), shape)
        if return_odf:
            pam.odf = np.reshape(np.array(pam.odf), shape)

        # Make sure all worker processes have exited before leaving context
        # manager in order to prevent temporary file deletion errors in windows
        pool.join()

    return pam


def _peaks_from_model_parallel_sub(args):
    (data_file_name, mask_file_name) = args[0]
    (start_pos, end_pos) = args[1]
    model = args[2]
    sphere = args[3]
    relative_peak_threshold = args[4]
    min_separation_angle = args[5]
    return_odf = args[6]
    return_sh = args[7]
    gfa_thr = args[8]
    normalize_peaks = args[9]
    sh_order = args[10]
    sh_basis_type = args[11]
    legacy = args[12]
    npeaks = args[13]
    B = args[14]
    invB = args[15]

    data = np.load(data_file_name, mmap_mode="r")[start_pos:end_pos]
    if mask_file_name is not None:
        mask = np.load(mask_file_name, mmap_mode="r")[start_pos:end_pos]
    else:
        mask = None

    return peaks_from_model(
        model,
        data,
        sphere,
        relative_peak_threshold,
        min_separation_angle,
        mask=mask,
        return_odf=return_odf,
        return_sh=return_sh,
        gfa_thr=gfa_thr,
        normalize_peaks=normalize_peaks,
        sh_order_max=sh_order,
        sh_basis_type=sh_basis_type,
        legacy=legacy,
        npeaks=npeaks,
        B=B,
        invB=invB,
        parallel=False,
        num_processes=None,
    )


@deprecated_params("sh_order", new_name="sh_order_max", since="1.9", until="2.0")
@warning_for_keywords()
def peaks_from_model(
    model,
    data,
    sphere,
    relative_peak_threshold,
    min_separation_angle,
    *,
    mask=None,
    return_odf=False,
    return_sh=True,
    gfa_thr=0,
    normalize_peaks=False,
    sh_order_max=8,
    sh_basis_type=None,
    legacy=True,
    npeaks=5,
    B=None,
    invB=None,
    parallel=False,
    num_processes=None,
):
    """Fit the model to data and computes peaks and metrics

    Parameters
    ----------
    model : a model instance
        `model` will be used to fit the data.
    data : ndarray
        Diffusion data.
    sphere : Sphere
        The Sphere providing discrete directions for evaluation.
    relative_peak_threshold : float
        Only return peaks greater than ``relative_peak_threshold * m`` where m
        is the largest peak.
    min_separation_angle : float in [0, 90] The minimum distance between
        directions. If two peaks are too close only the larger of the two is
        returned.
    mask : array, optional
        If `mask` is provided, voxels that are False in `mask` are skipped and
        no peaks are returned.
    return_odf : bool
        If True, the odfs are returned.
    return_sh : bool
        If True, the odf as spherical harmonics coefficients is returned
    gfa_thr : float
        Voxels with gfa less than `gfa_thr` are skipped, no peaks are returned.
    normalize_peaks : bool
        If true, all peak values are calculated relative to `max(odf)`.
    sh_order_max : int, optional
        Maximum SH order (l) in the SH fit.  For `sh_order_max`, there
        will be
        ``(sh_order_max + 1) * (sh_order_max + 2) / 2``
        SH coefficients (default 8).
    sh_basis_type : {None, 'tournier07', 'descoteaux07'}
        ``None`` for the default DIPY basis,
        ``tournier07`` for the Tournier 2007 :footcite:p:Tournier2007` basis,
        and ``descoteaux07`` for the Descoteaux 2007 :footcite:p:Descoteaux2007`
        basis
        (``None`` defaults to ``descoteaux07``).
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    npeaks : int
        Maximum number of peaks found (default 5 peaks).
    B : ndarray, optional
        Matrix that transforms spherical harmonics to spherical function
        ``sf = np.dot(sh, B)``.
    invB : ndarray, optional
        Inverse of B.
    parallel: bool
        If True, use multiprocessing to compute peaks and metric
        (default False). Temporary files are saved in the default temporary
        directory of the system. It can be changed using ``import tempfile``
        and ``tempfile.tempdir = '/path/to/tempdir'``.
    num_processes: int, optional
        If `parallel` is True, the number of subprocesses to use
        (default multiprocessing.cpu_count()). If < 0 the maximal number of
        cores minus ``num_processes + 1`` is used (enter -1 to use as many
        cores as possible). 0 raises an error.

    Returns
    -------
    pam : PeaksAndMetrics
        An object with ``gfa``, ``peak_directions``, ``peak_values``,
        ``peak_indices``, ``odf``, ``shm_coeffs`` as attributes

    References
    ----------
    .. footbibliography::

    """
    if return_sh and (B is None or invB is None):
        B, invB = sh_to_sf_matrix(
            sphere,
            sh_order_max=sh_order_max,
            basis_type=sh_basis_type,
            return_inv=True,
            legacy=legacy,
        )

    num_processes = determine_num_processes(num_processes)

    if parallel and num_processes > 1:
        # It is mandatory to provide B and invB to the parallel function.
        # Otherwise, a call to np.linalg.pinv is made in a subprocess and
        # makes it timeout on some system.
        # see https://github.com/dipy/dipy/issues/253 for details
        return _peaks_from_model_parallel(
            model,
            data,
            sphere,
            relative_peak_threshold,
            min_separation_angle,
            mask,
            return_odf,
            return_sh,
            gfa_thr,
            normalize_peaks,
            sh_order_max,
            sh_basis_type,
            legacy,
            npeaks,
            B,
            invB,
            num_processes,
        )

    shape = data.shape[:-1]
    if mask is None:
        mask = np.ones(shape, dtype="bool")
    else:
        if mask.shape != shape:
            raise ValueError("Mask is not the same shape as data.")

    gfa_array = np.zeros(shape)
    qa_array = np.zeros((shape + (npeaks,)))

    peak_dirs = np.zeros((shape + (npeaks, 3)))
    peak_values = np.zeros((shape + (npeaks,)))
    peak_indices = np.zeros((shape + (npeaks,)), dtype=np.int32)
    peak_indices.fill(-1)

    if return_sh:
        n_shm_coeff = (sh_order_max + 2) * (sh_order_max + 1) // 2
        shm_coeff = np.zeros((shape + (n_shm_coeff,)))

    if return_odf:
        odf_array = np.zeros((shape + (len(sphere.vertices),)))

    global_max = -np.inf
    for idx in ndindex(shape):
        if not mask[idx]:
            continue

        odf = model.fit(data[idx]).odf(sphere=sphere)

        if return_sh:
            shm_coeff[idx] = np.dot(odf, invB)

        if return_odf:
            odf_array[idx] = odf

        gfa_array[idx] = gfa(odf)
        if gfa_array[idx] < gfa_thr:
            global_max = max(global_max, odf.max())
            continue

        # Get peaks of odf
        direction, pk, ind = peak_directions(
            odf,
            sphere,
            relative_peak_threshold=relative_peak_threshold,
            min_separation_angle=min_separation_angle,
        )

        # Calculate peak metrics
        if pk.shape[0] != 0:
            global_max = max(global_max, pk[0])

            n = min(npeaks, pk.shape[0])
            qa_array[idx][:n] = pk[:n] - odf.min()

            peak_dirs[idx][:n] = direction[:n]
            peak_indices[idx][:n] = ind[:n]
            peak_values[idx][:n] = pk[:n]

            if normalize_peaks:
                peak_values[idx][:n] = peak_values[idx][:n] / pk[0] if pk[0] != 0 else 0
                peak_dirs[idx] *= peak_values[idx][:, None]

    qa_array /= global_max

    return _pam_from_attrs(
        PeaksAndMetrics,
        sphere,
        peak_indices,
        peak_values,
        peak_dirs,
        gfa_array,
        qa_array,
        shm_coeff if return_sh else None,
        B if return_sh else None,
        odf_array if return_odf else None,
    )


def reshape_peaks_for_visualization(peaks):
    """Reshape peaks for visualization.

    Reshape and convert to float32 a set of peaks for visualisation with mrtrix
    or the fibernavigator.

    Parameters
    ----------
    peaks: nd array (..., N, 3) or PeaksAndMetrics object
        The peaks to be reshaped and converted to float32.

    Returns
    -------
    peaks : nd array (..., 3*N)
    """
    if isinstance(peaks, PeaksAndMetrics):
        peaks = peaks.peak_dirs

    return peaks.reshape(np.append(peaks.shape[:-2], -1)).astype("float32")


def peaks_from_positions(
    positions,
    odfs,
    sphere,
    affine,
    *,
    pmf_gen=None,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    is_symmetric=True,
    npeaks=5,
):
    """
    Extract the peaks at each positions.

    Parameters
    ----------
    position : array, (N, 3)
        World coordinates of the N positions.
    odfs : array, (X, Y, Z, M)
        Orientation distribution function (spherical function) represented
        on a sphere of M points.
    sphere : Sphere
        A discrete Sphere. The M points on the sphere correspond to the points
        of the odfs.
    affine : array (4, 4)
        The mapping between voxel indices and the point space for positions.
    pmf_gen : PmfGen
        Probability mass function generator from voxel orientation information. Replaces
        ``odfs`` and ``sphere`` when used.
    relative_peak_threshold : float, optional
        Only peaks greater than ``min + relative_peak_threshold * scale`` are
        kept, where ``min = max(0, odf.min())`` and
        ``scale = odf.max() - min``. The ``relative_peak_threshold`` should
        be in the range [0, 1].
    min_separation_angle : float, optional
        The minimum distance between directions. If two peaks are too close
        only the larger of the two is returned. The ``min_separation_angle``
        should be in the range [0, 90].
    is_symmetric : bool, optional
        If True, v is considered equal to -v.
    npeaks : int, optional
        The maximum number of peaks to extract at from each position.

    Returns
    -------
    peaks_arr : array (N, npeaks, 3)
    """

    if pmf_gen is not None and (odfs is not None or sphere is not None):
        msg = (
            "``odfs`` and ``sphere`` arguments will be ignored in favor of ``pmf_gen``."
        )
        warnings.warn(msg, stacklevel=2)

    if pmf_gen is not None:
        # use the sphere data from the pmf_gen
        sphere = pmf_gen.get_sphere()

    inv_affine = np.linalg.inv(affine)
    vox_positions = np.dot(positions, inv_affine[:3, :3].T.copy())
    vox_positions += inv_affine[:3, 3]

    peaks_arr = np.zeros((len(positions), npeaks, 3))

    if vox_positions.dtype not in [np.float64, float]:
        vox_positions = vox_positions.astype(float)

    for i, s in enumerate(vox_positions):
        if pmf_gen:
            odf = pmf_gen.get_pmf(s)
        else:
            odf = trilinear_interpolate4d(odfs, s)
        peaks, _, _ = peak_directions(
            odf,
            sphere,
            relative_peak_threshold=relative_peak_threshold,
            min_separation_angle=min_separation_angle,
            is_symmetric=is_symmetric,
        )
        nbr_peaks = min(npeaks, peaks.shape[0])
        peaks_arr[i, :nbr_peaks, :] = peaks[:nbr_peaks, :]

    return peaks_arr
